#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains On-Line Time Warping.
"""

import numpy as np
import scipy
from enum import IntEnum
from queue import Queue
from scipy.spatial.distance import cdist


def element_of_metric(vec1, vec2):
    """
    metric that evaluates occurence of vec2 (scalar) in vec1 (vector n-dim)
    """
    return 2 - np.sum(vec2 == vec1)


def cdist_local(arr1, arr2, metric):
    """
    compute array of pairwise distances between
    the elements of two arrays given a metric

    Parameters
    ----------
    arr1: numpy nd array

    arr2: numpy nd array

    metric: callable
        a metric function

    Returns
    -------

    pdist_array: numpy 2d array
        array of pairwise distances
    """
    pdist_array = np.ones((arr1.shape[0], arr2.shape[0])) * np.inf
    for i in range(arr1.shape[0]):
        for j in range(arr2.shape[0]):
            pdist_array[i, j] = metric(arr1[i], arr2[j])
    return pdist_array


class Direction(IntEnum):
    REF = 0
    INPUT = 1
    BOTH = 2

    def toggle(self):
        return Direction(self ^ 1) if self != Direction.BOTH else Direction.INPUT


class OLTW(object):
    """
    On-line Dynamic Time Warping (OLTW) algorithm for aligning a input sequence to a reference sequence.
    Adapted from: https://github.com/laurenceyoon/python-match

    Parameters
    ----------
    reference_features : np.ndarray
        A 2D array with dimensions (`T(n_timesteps)`, `F(n_features)`) containing the
        features of the reference the input is going to be aligned to.

    queue : Queue
        A queue for storing the input features, which shares the same object as the audio stream.

    window_size : int
        Size of the window (in steps) of the cost matrix.

    max_run_count : int
        Maximum number of times the class can run in the same direction.

    hop_size : int
        number of seq items that get added at step

    directional_weights: np.ndarray
        weights associated with each of the three possible steps

    Attributes
    ----------
    warping_path : np.ndarray [shape=(2, T)]
        Warping path with pairs of indices of the reference and input features.
        where warping_path[0] is the index of the reference feature and
        warping_path[1] is the index of the input feature.
    """

    def __init__(
        self,
        reference_features=None,
        input_feature_shape=None,
        queue=None,
        window_size=10,  # shape of the acc cost matric
        max_run_count=100,  # maximal number of steps
        hop_size=1,  # number of seq items that get added at step
        directional_weights=np.array([1, 1, 1]),  # down, diag, right
        cdist_fun=cdist,
        cdist_metric="euclidean",
        **kwargs,
    ):
        self.queue = queue
        self.cdist_fun = cdist_fun
        self.cdist_metric = cdist_metric
        self.directional_weights = directional_weights
        if reference_features is not None:
            self.set_feature_arrays(reference_features, input_feature_shape)
        else:
            self.reference_features = None
            self.N_ref = None
            self.input_features = None
        self.w = window_size
        self.max_run_count = max_run_count
        self.hop_size = hop_size
        self.initialize()

    def set_feature_arrays(
        self,
        reference_features,
        input_feature_shape=None,
    ):
        self.reference_features = reference_features
        self.N_ref = self.reference_features.shape[0]
        if input_feature_shape is not None:
            self.input_features = np.empty((0, input_feature_shape))
        else:
            self.input_features = np.empty((0, self.reference_features.shape[1]))

    def initialize(self):
        self.ref_pointer = 0
        self.input_pointer = 0
        self.run_count = 0
        self.previous_direction = None
        self.wp = np.array([[-1, -1]]).T  # [shape=(2, T)]

        self.ref_pointer += self.w  # window of ref shifted at start
        self.acc_dist_matrix = np.full((self.w, self.w), np.inf)
        self.acc_dist_matrix[0, :] = 0
        self.acc_len_matrix = np.zeros((self.w, self.w))
        self.queue_non_empty = True
        self.local_both_dist = np.zeros((self.w, self.w))

    @property
    def warping_path(self):  # [shape=(2, T)]
        return self.wp[:, 1:]

    def offset(self):
        offset_x = max(self.ref_pointer - self.w, 0)
        offset_y = max(self.input_pointer - self.w, 0)
        return np.array([offset_x, offset_y])

    def update_ref_direction(self, dist, new_acc, new_len_acc, wx, wy, d):
        update_x0 = wx - d
        for i in range(d):
            for j in range(wy):
                local_dist = dist[i, j]
                if j == 0:
                    new_acc[update_x0 + i, j] = (
                        local_dist * self.directional_weights[0]
                        + new_acc[update_x0 + i - 1, j]
                    )
                    new_len_acc[update_x0 + i, j] = (
                        1 + new_len_acc[update_x0 + i - 1, j]
                    )
                else:
                    compares = [
                        new_acc[update_x0 + i - 1, j]
                        + local_dist * self.directional_weights[0],
                        new_acc[update_x0 + i, j - 1]
                        + local_dist * self.directional_weights[2],
                        new_acc[update_x0 + i - 1, j - 1]
                        + local_dist * self.directional_weights[1],  # diagonal
                    ]
                    len_compares = [
                        new_len_acc[update_x0 + i - 1, j],
                        new_len_acc[update_x0 + i, j - 1],
                        new_len_acc[update_x0 + i - 1, j - 1],
                    ]
                    local_direction = np.argmin(compares)
                    new_acc[update_x0 + i, j] = compares[local_direction]
                    new_len_acc[update_x0 + i, j] = 1 + len_compares[local_direction]
        return new_acc, new_len_acc

    def update_input_direction(self, dist, new_acc, new_len_acc, wx, wy, d):
        update_y0 = wy - d
        for i in range(wx):
            for j in range(d):
                local_dist = dist[i, j]
                if i == 0:
                    new_acc[i, update_y0 + j] = (
                        local_dist * self.directional_weights[2]
                        + new_acc[i, update_y0 - 1 + j]
                    )
                    new_len_acc[i, update_y0 + j] = (
                        1 + new_len_acc[i, update_y0 - 1 + j]
                    )
                else:
                    compares = [
                        new_acc[i - 1, update_y0 + j]
                        + local_dist * self.directional_weights[0],
                        new_acc[i, update_y0 + j - 1]
                        + local_dist * self.directional_weights[2],
                        new_acc[i - 1, update_y0 + j - 1]
                        + local_dist * self.directional_weights[1],  # diagonal
                    ]
                    len_compares = [
                        new_len_acc[i - 1, update_y0 + j],
                        new_len_acc[i, update_y0 + j - 1],
                        new_len_acc[i - 1, update_y0 + j - 1],
                    ]
                    local_direction = np.argmin(compares)
                    new_acc[i, update_y0 + j] = compares[local_direction]
                    new_len_acc[i, update_y0 + j] = 1 + len_compares[local_direction]
        return new_acc, new_len_acc

    def update_both_direction(self, dist, new_acc, new_len_acc, wx, wy, d):
        for i in range(wx):
            for j in range(wy):
                local_dist = dist[i, j]
                if i == 0 and j >= wy - d:
                    new_acc[i, j] = (
                        local_dist * self.directional_weights[2] + new_acc[i, j - 1]
                    )
                    new_len_acc[i, j] = 1 + new_len_acc[i, j - 1]
                elif i >= wx - d and j == 0:
                    new_acc[i, j] = (
                        local_dist * self.directional_weights[0] + new_acc[i - 1, j]
                    )
                    new_len_acc[i, j] = 1 + new_len_acc[i - 1, j]
                elif i >= wx - d or j >= wy - d:
                    compares = [
                        new_acc[i - 1, j] + local_dist * self.directional_weights[0],
                        new_acc[i, j - 1] + local_dist * self.directional_weights[2],
                        new_acc[i - 1, j - 1]
                        + local_dist * self.directional_weights[1],
                    ]
                    len_compares = [
                        new_len_acc[i - 1, j],
                        new_len_acc[i, j - 1],
                        new_len_acc[i - 1, j - 1],
                    ]
                    local_direction = np.argmin(compares)
                    new_acc[i, j] = compares[local_direction]
                    new_len_acc[i, j] = 1 + len_compares[local_direction]
        return new_acc, new_len_acc

    def update_both_direction_new(self, dist, new_acc, new_len_acc, wx, wy, d):
        for j in range(wy - d, wy):
            new_acc[0, j] = dist[0, j] * self.directional_weights[2] + new_acc[0, j - 1]
            new_len_acc[0, j] = 1 + new_len_acc[0, j - 1]

        for i in range(wx - d, wx):
            new_acc[i, 0] = dist[i, 0] * self.directional_weights[0] + new_acc[i - 1, 0]
            new_len_acc[i, 0] = 1 + new_len_acc[i - 1, 0]

        # first add the columns to the right
        # do not go all the way to the corner

        for i in range(1, wx - d):
            for j in range(wy - d, wy):
                local_dist = dist[i, j]
                compares = [
                    new_acc[i - 1, j] + local_dist * self.directional_weights[0],
                    new_acc[i, j - 1] + local_dist * self.directional_weights[2],
                    new_acc[i - 1, j - 1] + local_dist * self.directional_weights[1],
                ]
                len_compares = [
                    new_len_acc[i - 1, j],
                    new_len_acc[i, j - 1],
                    new_len_acc[i - 1, j - 1],
                ]
                local_direction = np.argmin(compares)
                new_acc[i, j] = compares[local_direction]
                new_len_acc[i, j] = 1 + len_compares[local_direction]

        # then add the rows at the bottom
        for i in range(wx - d, wx):
            for j in range(1, wy):
                local_dist = dist[i, j]
                compares = [
                    new_acc[i - 1, j] + local_dist * self.directional_weights[0],
                    new_acc[i, j - 1] + local_dist * self.directional_weights[2],
                    new_acc[i - 1, j - 1] + local_dist * self.directional_weights[1],
                ]
                len_compares = [
                    new_len_acc[i - 1, j],
                    new_len_acc[i, j - 1],
                    new_len_acc[i - 1, j - 1],
                ]
                local_direction = np.argmin(compares)
                new_acc[i, j] = compares[local_direction]
                new_len_acc[i, j] = 1 + len_compares[local_direction]

        return new_acc, new_len_acc

    def update_cost_matrix(self, direction):
        # local cost matrix
        x, y = self.ref_pointer, self.input_pointer
        wx, wy = min(self.w, x), min(self.w, y)
        d = self.hop_size
        new_acc = np.full((wx, wy), np.inf, dtype=np.float32)
        # new_acc[0,0] = 0
        new_len_acc = np.zeros((wx, wy))

        if direction is Direction.REF:
            new_acc[:-d, :] = self.acc_dist_matrix[d:]
            new_len_acc[:-d, :] = self.acc_len_matrix[d:]
            x_seg = self.reference_features[x - d : x]  # [d, 12]
            y_seg = self.input_features[y - wy : y]  # [wy, 12]
            dist = self.cdist_fun(x_seg, y_seg, metric=self.cdist_metric)  # [wx, d]
            new_acc, new_len_acc = self.update_ref_direction(
                dist, new_acc, new_len_acc, wx, wy, d
            )

        elif direction is Direction.INPUT:
            overlap_y = wy - d
            new_acc[:, :-d] = self.acc_dist_matrix[:, -overlap_y:]
            new_len_acc[:, :-d] = self.acc_len_matrix[:, -overlap_y:]
            x_seg = self.reference_features[x - wx : x]  # [wx, 12]
            y_seg = self.input_features[y - d : y]  # [d, 12]
            dist = self.cdist_fun(x_seg, y_seg, metric=self.cdist_metric)  # [wx, d]
            new_acc, new_len_acc = self.update_input_direction(
                dist, new_acc, new_len_acc, wx, wy, d
            )

        elif direction is Direction.BOTH:  # input, BOTH
            overlap_y = wy - d
            new_acc[:-d, :-d] = self.acc_dist_matrix[d:, -overlap_y:]
            new_len_acc[:-d, :-d] = self.acc_len_matrix[d:, -overlap_y:]

            # old costly distance computation
            # x_seg = self.reference_features[x - wx : x]  # [wx, 12]
            # y_seg = self.input_features[y - wy : y]  # [wy, 12]
            # dist = self.cdist_fun(x_seg, y_seg, metric=self.cdist_metric)  # [wx, d]
            # new, compute only outer slice
            x_seg = self.reference_features[x - wx : x - d]  # [wx - d]
            y_seg = self.input_features[y - d : y]  # [d]
            self.local_both_dist[: wx - d, -d:] = self.cdist_fun(
                x_seg, y_seg, metric=self.cdist_metric
            )
            x_seg = self.reference_features[x - d : x]  # [d]
            y_seg = self.input_features[y - wy : y]  # [wy]
            self.local_both_dist[wx - d :, :] = self.cdist_fun(
                x_seg, y_seg, metric=self.cdist_metric
            )

            new_acc, new_len_acc = self.update_both_direction_new(
                self.local_both_dist, new_acc, new_len_acc, wx, wy, d
            )
        self.acc_dist_matrix = new_acc
        self.acc_len_matrix = new_len_acc

    def first_cost_matrix(self):
        # local cost matrix
        x, y = self.ref_pointer, self.input_pointer  # should be window, hop_size
        wx, wy = x, y + 1  # append an extra column
        d = self.hop_size
        new_acc = np.full((wx, wy), np.inf, dtype=np.float32)
        new_acc[0, 0] = 0  # starting point in corner
        new_len_acc = np.zeros((wx, wy))
        x_seg = self.reference_features[x - wx : x]  # [wx, 12]
        y_seg = self.input_features[y - d : y]  # [d, 12]
        dist = self.cdist_fun(x_seg, y_seg, metric=self.cdist_metric)  # [wx, d]
        new_acc, new_len_acc = self.update_input_direction(
            dist, new_acc, new_len_acc, wx, wy, d
        )

        self.acc_dist_matrix = new_acc[:, 1:]
        self.acc_len_matrix = new_len_acc[:, 1:]

    def select_candidate(self):
        norm_x_edge = self.acc_dist_matrix[-1, :] / self.acc_len_matrix[-1, :]
        norm_y_edge = self.acc_dist_matrix[:, -1] / self.acc_len_matrix[:, -1]
        cat = np.concatenate((norm_x_edge, norm_y_edge))

        min_idx = np.argmin(cat)
        offset = self.offset()
        if min_idx < len(norm_x_edge):
            self.candidate = np.array([self.ref_pointer - offset[0] - 1, min_idx])
        else:
            self.candidate = np.array(
                [min_idx - len(norm_x_edge), self.input_pointer - offset[1] - 1]
            )

        # print("edge",cat)
        # print("pointers:", self.ref_pointer, self.input_pointer, offset)
        # print("cand",self.candidate, min_idx)

    def add_candidate_to_path(self):
        new_coordinate = np.expand_dims(
            self.offset() + self.candidate, axis=1
        )  # [2, 1]
        self.wp = np.concatenate((self.wp, new_coordinate), axis=1)

    def select_next_direction(self):
        if self.input_pointer < self.w:
            next_direction = Direction.INPUT
        elif self.run_count > self.max_run_count:
            next_direction = self.previous_direction.toggle()
        else:
            offset = self.offset()
            x0, y0 = offset[0], offset[1]
            ref_bool = self.candidate[0] == self.ref_pointer - x0 - 1
            input_bool = self.candidate[1] == self.input_pointer - y0 - 1
            if ref_bool and input_bool:
                next_direction = Direction.BOTH
            elif ref_bool:
                next_direction = Direction.REF
            else:
                assert input_bool
                next_direction = Direction.INPUT
        return next_direction

    def get_new_input(self):
        try:
            input_feature = self.queue.get(block=False)
            self.input_features = np.vstack([self.input_features, input_feature])
            self.input_pointer += self.hop_size

        except:
            print("empty queue")
            self.queue_non_empty = False

    def is_still_following(self):
        is_still_following = self.ref_pointer <= (self.N_ref - self.hop_size)
        return is_still_following and self.queue_non_empty

    def handle_direction(self, direction):
        # get fitting features/windows
        if direction is not Direction.REF:
            self.get_new_input()
        if direction is not Direction.INPUT:
            self.ref_pointer += self.hop_size

        # accumulate run count
        if direction == self.previous_direction:
            self.run_count += 1
        else:
            self.run_count = 1
        self.previous_direction = direction

    def handle_first_input(self):
        direction = self.select_next_direction()
        self.handle_direction(direction)
        self.first_cost_matrix()
        self.select_candidate()
        self.add_candidate_to_path()

    def run(self, verbose=False):
        print("Start running OLTW")
        self.initialize()
        self.handle_first_input()
        while self.is_still_following():
            direction = self.select_next_direction()
            if verbose:
                print("ACC DIST \n", self.acc_dist_matrix)
                print("ACC LEN \n", self.acc_len_matrix)
                print("RECENT WARPING PATH \n", self.warping_path[:, -3:])
                print("NEXT DIRECTION \n", direction)
                print("*" * 50)
            print("RECENT WARPING PATH: ", self.warping_path[:, -1:])
            self.handle_direction(direction)
            self.update_cost_matrix(direction)
            self.select_candidate()
            self.add_candidate_to_path()
        print("... and we're done.")
        return self.warping_path


if __name__ == "__main__":
    RANGE_L = 6
    REPEATS = 3
    HOP_SIZE = 1
    WINDOW_SIZE = 3
    # r = np.row_stack([np.arange(RANGE_L).reshape(-1,1) for k in range(REPEATS)]).astype(float)
    # t = np.row_stack([np.roll(np.arange(RANGE_L),0,0).reshape(-1,1) for k in range(REPEATS)]).astype(float)
    # t +=0.5
    # r = np.arange(40).reshape(-1,4)
    # t = np.arange(40).reshape(-1,1)

    r = np.row_stack([np.arange(3).reshape(-1, 3) for k in range(10)])
    t = np.row_stack([np.arange(3).reshape(-1, 1) for k in range(10)])

    queue = Queue()
    # for tt in range((RANGE_L * REPEATS) // HOP_SIZE):
    #     queue.put(t[tt*HOP_SIZE: (tt+1)*HOP_SIZE,:])
    for tt in range(40):
        queue.put(t[tt : tt + 1, :])

    o = OLTW(
        reference_features=r,
        input_feature_shape=1,
        queue=queue,
        frame_per_seg=HOP_SIZE,
        window_size=WINDOW_SIZE,
        max_run_count=4,
        directional_weights=np.array([1.05, 1.1, 1]),
        cdist_fun=cdist_local,
        cdist_metric=element_of_metric,
    )
    p = o.run()
    print("path \n", p)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
The top level of the package contains functions to
note-align symbolic music data.
"""
import pkg_resources

EXAMPLE = pkg_resources.resource_filename("parangonar", "assets/mozart_k265_var1.match")
ALIGNMENT_TRANSFORMER_CHECKPOINT = pkg_resources.resource_filename(
    "parangonar", "assets/alignment_transformer_checkpoint.pt"
)
THEGLUENOTE_CHECKPOINT = pkg_resources.resource_filename(
    "parangonar", "assets/thegluenote_small_checkpoint.pt"
)
from .match import (
    AnchorPointNoteMatcher,
    AutomaticNoteMatcher,
    DualDTWNoteMatcher,
    TheGlueNoteMatcher,
)
from .match import OnlineTransformerMatcher, OnlinePureTransformerMatcher
from .evaluate import (
    fscore_alignments,
    print_fscore_alignments,
    plot_alignment,
    plot_alignment_comparison,
)

__all__ = [
    "AnchorPointNoteMatcher",
    "AutomaticNoteMatcher",
    "DualDTWNoteMatcher",
    "TheGlueNoteMatcher",
    "OnlineTransformerMatcher",
    "OnlinePureTransformerMatcher" "fscore_alignments",
    "plot_alignment_comparison",
    "plot_alignment",
]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This module includes tests for alignment utilities.
"""
import unittest
import numpy as np
from parangonar.dp.dtw import DTW, WDTW, DTWSL, FDTW, l2
from parangonar.dp.nwtw import NW_DTW, NW, WNWTW, ONW

RNG = np.random.RandomState(1984)

array1 = np.array([[0, 1, 2, 3, 6]]).T
array2 = np.array([[0, 1, 2, 3, 4, 5, 6]]).T

result_dtw = np.array([[0, 0], [1, 1], [2, 2], [3, 3], [3, 4], [4, 5], [4, 6]])

result_nw = np.array([[0, 0], [1, 1], [2, 2], [3, 3], [-1, 4], [-1, 5], [4, 6]])

result_wnw = np.array([[0, 0], [1, 1], [2, 2], [3, 3], [3, 4], [3, 5], [4, 6]])


class TestDTWAlignment(unittest.TestCase):
    def test_DTW_align(self, **kwargs):
        vanillaDTW = DTW()
        _, path = vanillaDTW(array1, array2, return_path=True)
        self.assertTrue(np.all(result_dtw == path))

    def test_DTWSL_align(self, **kwargs):
        vanillaDTWSL = DTWSL(metric=l2)
        _, path = vanillaDTWSL(array1, array2, return_path=True)
        self.assertTrue(np.all(result_dtw == path))

    def test_WDTW_align(self, **kwargs):
        weightedDTW = WDTW()
        path = weightedDTW(array1, array2)[0]
        self.assertTrue(np.all(result_dtw == path))

    def test_FDTW_align(self, **kwargs):
        FlexDTW = FDTW()
        path = FlexDTW(array1, array2)[0]
        self.assertTrue(np.all(result_dtw == path))


class TestNWDTWAlignment(unittest.TestCase):
    def test_NWDTW_align(self, **kwargs):
        vanillaNW_DTW = NW_DTW()
        _, path = vanillaNW_DTW(array1, array2)
        self.assertTrue(np.all(result_nw == path))

    def test_NW_align(self, **kwargs):
        vanillaNW = NW()
        _, path = vanillaNW(array1, array2)
        self.assertTrue(np.all(result_nw == path))

    def test_WNWTW_align(self, **kwargs):
        weightedNW = WNWTW()
        path = weightedNW(array1, array2)[0]
        self.assertTrue(np.all(result_wnw == path))

    def test_ONW_align(self, **kwargs):
        originalNW = ONW()
        path = originalNW(array1, array2)[0]
        self.assertTrue(np.all(result_wnw == path))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains preprocessing methods
"""

import numpy as np
from scipy.interpolate import interp1d

from partitura.utils.music import compute_pianoroll

from ..dp.dtw import DTW
from ..dp.nwtw import NW_DTW, NW


################################### HELPERS ###################################


def alignment_times_from_dtw(
    score_note_array,
    performance_note_array,
    matcher=DTW(),
    SCORE_FINE_NODE_LENGTH=1.0,
    s_time_div=16,
    p_time_div=16,
):
    """

    Coarse time warping to generate anchor points
    Returns an array, each row is tuple of corresponding
    times in the score array and the perfromance array,
    respectively.

    Args:
        score_note_array (_type_): _description_
        performance_note_array (_type_): _description_
        matcher (_type_, optional): _description_. Defaults to DTW().
        SCORE_FINE_NODE_LENGTH (float, optional): _description_. Defaults to 1.0.
        s_time_div (int, optional): _description_. Defaults to 16.
        p_time_div (int, optional): _description_. Defaults to 16.

    Returns:
        _type_: _description_
    """
    # _____________ fine alignment ____________
    # compute proper piano rolls
    s_pianoroll = compute_pianoroll(
        score_note_array, time_div=s_time_div, remove_drums=False
    ).toarray()
    p_pianoroll = compute_pianoroll(
        performance_note_array, time_div=p_time_div, remove_drums=False
    ).toarray()
    # make piano rolls binary
    p_pianoroll_ones = np.zeros_like(p_pianoroll)
    p_pianoroll_ones[p_pianoroll > 0.0] = 1.0
    # align the piano rolls
    _, path = matcher(s_pianoroll.T, p_pianoroll_ones.T)
    # compute an alignment of times using the DTW path
    path_array = np.array(path)

    # Post-process path for NW-DTW (insertion and deletions
    # are denoted as -1 in the path)
    if isinstance(matcher, (NW_DTW, NW)):
        valid_idxs = np.where(
            np.logical_and(path_array[:, 0] != -1, path_array[:, 1] != -1)
        )[0]
        path_array = path_array[valid_idxs]

        if len(path_array) == 0:
            path_array = np.array(
                [
                    (path[:, 0].min(), path[:, 1].min()),
                    (path[:, 1].max(), path[:, 1].max()),
                ],
                dtype=int,
            )

    times_score = path_array[:, 0] / s_time_div
    times_performance = path_array[:, 1] / p_time_div

    u_times_score = np.unique(times_score)
    u_times_score_idxs = [np.where(times_score == u)[0] for u in u_times_score]
    u_times_performance = np.array(
        [np.min(times_performance[ui]) for ui in u_times_score_idxs]
    )

    try:
        # Use a mapping to deal with missing values (due to
        # insertions and deletions in NW-related methods)
        # It should not affect the behavior of DTW methods
        # CC: I will check this just in case ;)
        stime_to_ptime_map = interp1d(
            x=u_times_score,
            y=u_times_performance,
            kind="linear",
            bounds_error=False,
            fill_value="extrapolate",
        )
    except ValueError:
        # if there is only one element in u_times_score
        stime_to_ptime_map = (
            lambda x: np.ones_like(x) * u_times_performance[0]
        )  # noqa: E731

    min_score = times_score.min()
    max_score = max(
        score_note_array["onset_beat"].max() - score_note_array["onset_beat"].min(),
        SCORE_FINE_NODE_LENGTH,
    )
    max_performance = max(
        performance_note_array["onset_sec"].max()
        - performance_note_array["onset_sec"].min(),
        SCORE_FINE_NODE_LENGTH,
    )

    cut_times_score = np.r_[
        np.arange(min_score, max_score, SCORE_FINE_NODE_LENGTH), max_score
    ]

    cut_times_performance = np.clip(
        stime_to_ptime_map(cut_times_score), a_min=0, a_max=max_performance
    )

    cut_times_performance += performance_note_array["onset_sec"].min()
    cut_times_score += score_note_array["onset_beat"].min()

    dtw_alignment_times = np.column_stack((cut_times_score, cut_times_performance))
    return dtw_alignment_times


def traverse_the_alignment_graph(
    start_id,
    score_ids,
    performance_ids,
    performance_alignment,
    score_alignment,
    counter,
    max_depth=150,
):
    # score_ids = [input_id]
    # performance_id = []
    if start_id not in score_ids and counter < max_depth:
        # print("appendin", start_id)
        score_ids.append(start_id)
        new_pids = score_alignment[start_id]
        for pid in new_pids:
            # print("loop point", pid)
            if pid not in performance_ids:
                # print("appendin perf", pid)
                performance_ids.append(pid)
                new_sids = performance_alignment[pid]
                for sid in new_sids:
                    counter += 1
                    traverse_the_alignment_graph(
                        sid,
                        score_ids,
                        performance_ids,
                        performance_alignment,
                        score_alignment,
                        counter,
                        max_depth,
                    )

            else:
                continue
    elif counter == 150:
        print("max recursion depth in note graph")
        pass
    else:
        pass
        # print("done")


################################### SEGMENT CUTTING ###################################


def cut_note_arrays(
    performance_note_array,
    score_note_array,
    alignment,
    sfuzziness=0.0,
    pfuzziness=0.0,
    window_size=1,
    pfuzziness_relative_to_tempo=False,
):
    """
    cut note arrays into two lists of corresponding
    note array segments based on anchor points given
    as "alignment": n by 2 array, first column score,
    second performace times.


    Args:
        performance_note_array (_type_): _description_
        score_note_array (_type_): _description_
        alignment (_type_): _description_
        sfuzziness (float, optional): _description_. Defaults to 0.0.
        pfuzziness (float, optional): _description_. Defaults to 0.0.
        window_size (int, optional): _description_. Defaults to 1.
        pfuzziness_relative_to_tempo (bool, optional): _description_. Defaults to False.

    Returns:
        _type_: _description_
    """

    score_note_arrays = []
    performance_note_arrays = []

    if not pfuzziness_relative_to_tempo:
        local_pfuzzines = np.ones_like(alignment[:, 1]) * pfuzziness
    else:
        approximate_tempo = interp1d(
            alignment[:-1, 0],
            np.diff(alignment[:, 1]) / np.diff(alignment[:, 0]),
            fill_value="extrapolate",
        )
        local_pfuzzines = approximate_tempo(alignment[:, 0]) * pfuzziness

    for i in range(len(alignment) - window_size):
        # extract all score notes with onsets inside the closed inter beat interval
        score_window_mask = np.all(
            [
                score_note_array["onset_beat"] >= alignment[i, 0] - sfuzziness,
                score_note_array["onset_beat"]
                < alignment[i + window_size, 0] + sfuzziness,
            ],
            axis=0,
        )
        # extract all performance notes with onsets inside the inter beat interval plus some fuzzy relaxation
        performance_window_mask = np.all(
            [
                performance_note_array["onset_sec"]
                >= alignment[i, 1] - local_pfuzzines[i],
                performance_note_array["onset_sec"]
                < alignment[i + window_size, 1] + local_pfuzzines[i],
            ],
            axis=0,
        )
        score_note_arrays.append(score_note_array[score_window_mask])
        performance_note_arrays.append(performance_note_array[performance_window_mask])

    return score_note_arrays, performance_note_arrays


################################### SEGMENT MENDING ###################################


def mend_note_alignments(
    note_alignments,
    performance_note_array,
    score_note_array,
    node_times,
    symbolic_note_matcher,
    max_traversal_depth=150,
):
    """
    mend note alignments in (overlapping) windows.
    creates a global dictionary of MAPS style alignments
    from a list of windowed MAPS style alignments
    """

    score_alignment = {"insertion": []}
    performance_alignment = {"deletion": []}
    alignment = []

    # approximate_position = interp1d(node_times[:,0],node_times[:,1],fill_value="extrapolate")
    # approximate_tempo = interp1d(node_times[:-1,0],np.diff(node_times[:,1])/np.diff(node_times[:,0]),fill_value="extrapolate")

    # score_onsets_by_id = { snote["id"]:snote["onset_beat"] for snote in score_note_array}

    # combine all note alignments in two dictionaries
    for window_id in range(len(note_alignments)):
        for alignment_line in note_alignments[window_id]:
            if alignment_line["label"] == "match":
                if alignment_line["score_id"] in score_alignment.keys():
                    score_alignment[alignment_line["score_id"]].append(
                        alignment_line["performance_id"]
                    )
                else:
                    score_alignment[alignment_line["score_id"]] = [
                        alignment_line["performance_id"]
                    ]

                if alignment_line["performance_id"] in performance_alignment.keys():
                    performance_alignment[alignment_line["performance_id"]].append(
                        alignment_line["score_id"]
                    )
                else:
                    performance_alignment[alignment_line["performance_id"]] = [
                        alignment_line["score_id"]
                    ]

            if alignment_line["label"] == "deletion":
                if alignment_line["score_id"] in score_alignment.keys():
                    score_alignment[alignment_line["score_id"]].append("deletion")
                else:
                    score_alignment[alignment_line["score_id"]] = ["deletion"]

            if alignment_line["label"] == "insertion":
                if alignment_line["performance_id"] in performance_alignment.keys():
                    performance_alignment[alignment_line["performance_id"]].append(
                        "insertion"
                    )
                else:
                    performance_alignment[alignment_line["performance_id"]] = [
                        "insertion"
                    ]

    # compute as single unique ids alignment
    used_perf_notes = ["deletion"]
    used_score_notes = ["insertion"]
    for snote_id in score_alignment.keys():
        # IF 1 p_note in alignment:
        if snote_id not in used_score_notes:
            if len(score_alignment[snote_id]) == 1:
                # DELETION
                if score_alignment[snote_id][0] == "deletion":
                    alignment.append({"label": "deletion", "score_id": snote_id})
                    used_score_notes.append(snote_id)
                else:
                    score_ids = []
                    performance_ids = []
                    counter = 0
                    traverse_the_alignment_graph(
                        snote_id,
                        score_ids,
                        performance_ids,
                        performance_alignment,
                        score_alignment,
                        counter,
                        max_depth=max_traversal_depth,
                    )
                    # deal with the possible alignments
                    no_insertion = [sid for sid in score_ids if sid != "insertion"]
                    no_deletion = [pid for pid in performance_ids if pid != "deletion"]
                    # if two unique notes match
                    if len(no_deletion) == 1 and len(no_insertion) == 1:
                        alignment.append(
                            {
                                "label": "match",
                                "score_id": snote_id,
                                "performance_id": str(score_alignment[snote_id][0]),
                            }
                        )
                        used_score_notes.append(no_insertion[0])
                        used_perf_notes.append(str(no_deletion[0]))
                    # try realigning
                    else:
                        local_score_note_array = score_note_array[
                            np.any(
                                [
                                    score_note_array["id"] == str(idx)
                                    for idx in no_insertion
                                ],
                                axis=0,
                            )
                        ]
                        masks = [
                            performance_note_array["id"] == str(idx)
                            for idx in no_deletion
                        ]
                        local_performance_note_array = performance_note_array[
                            np.any(masks, axis=0)
                        ]
                        # print( snote_id, "realigning with score and performance array: ",local_score_note_array.shape, local_performance_note_array.shape)

                        local_alignment = symbolic_note_matcher(
                            local_score_note_array,
                            local_performance_note_array,
                            node_times,
                        )
                        # import pdb; pdb.set_trace()
                        used_score_notes += no_insertion
                        used_perf_notes += no_deletion
                        alignment += local_alignment

            else:
                # multiple perf_ids aligned to snote_id
                score_ids = []
                performance_ids = []
                counter = 0
                traverse_the_alignment_graph(
                    snote_id,
                    score_ids,
                    performance_ids,
                    performance_alignment,
                    score_alignment,
                    counter,
                    max_depth=max_traversal_depth,
                )

                # deal with the possible alignments
                no_insertion = [sid for sid in score_ids if sid != "insertion"]
                no_deletion = [pid for pid in performance_ids if pid != "deletion"]
                # if two unique notes match
                if len(no_deletion) == 1 and len(no_insertion) == 1:
                    alignment.append(
                        {
                            "label": "match",
                            "score_id": snote_id,
                            "performance_id": str(no_deletion[0]),
                        }
                    )
                    used_score_notes.append(no_insertion[0])
                    used_perf_notes.append(str(no_deletion[0]))

                elif len(no_deletion) == 0:
                    # all deletions
                    alignment.append({"label": "deletion", "score_id": snote_id})
                    used_score_notes.append(snote_id)
                else:
                    # try realigning

                    local_score_note_array = score_note_array[
                        np.any(
                            [
                                score_note_array["id"] == str(idx)
                                for idx in no_insertion
                            ],
                            axis=0,
                        )
                    ]
                    local_performance_note_array = performance_note_array[
                        np.any(
                            [
                                performance_note_array["id"] == str(idx)
                                for idx in no_deletion
                            ],
                            axis=0,
                        )
                    ]

                    # print( snote_id,"realigning (conflicting) with score and performance array: ",local_score_note_array.shape, local_performance_note_array.shape)
                    local_alignment = symbolic_note_matcher(
                        local_score_note_array, local_performance_note_array, node_times
                    )

                    # import pdb; pdb.set_trace()
                    used_score_notes += no_insertion
                    used_perf_notes += no_deletion
                    alignment += local_alignment

    for pnote_id in performance_alignment.keys():
        if pnote_id not in used_perf_notes:
            # print("previously unused performance note: ", pnote_id, performance_alignment[pnote_id])
            alignment.append({"label": "insertion", "performance_id": pnote_id})

    return alignment, score_alignment, performance_alignment


################################### NOTE ARRAY ###################################


def note_per_ons_encoding(sarray):
    """
    create an onset-wise pitch based encoding
    of score note arrays
    """
    score_onsets = np.unique(sarray["onset_beat"])

    how_many_notes_per_ons = list()
    notes_per_ons = dict()

    # get unique pitches per score onset
    for ons in score_onsets:
        notes_per_ons[ons] = np.unique(sarray[sarray["onset_beat"] == ons]["pitch"])
        how_many_notes_per_ons.append(notes_per_ons[ons].shape[0])

    how_many_notes_per_ons = np.array(how_many_notes_per_ons)
    max_notes_per_ons = np.max(how_many_notes_per_ons)
    score_note_per_ons_encoding = (
        np.ones((score_onsets.shape[0], max_notes_per_ons)) * -1
    )

    # encode unique pitches per onset in array
    for idx, ons in enumerate(score_onsets):
        notes = notes_per_ons[ons]
        score_note_per_ons_encoding[idx, : notes.shape[0]] = notes

    return score_note_per_ons_encoding



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains pretrained pytorch models.
checkpoints in assets. 
loaded in matchers and online_matchers.
"""

import torch
import torch.nn as nn
import numpy as np
from torch.nn import TransformerEncoder, TransformerEncoderLayer
from torch.nn.modules.normalization import LayerNorm
import torch.nn.functional as F

# ALIGNMENT TRANSFORMER


class PositionalEncoding(nn.Module):
    def __init__(self, dim_model, dropout_p, max_len):
        super().__init__()

        self.dropout = nn.Dropout(dropout_p)
        pos_encoding = torch.zeros(max_len, dim_model)
        positions_list = torch.arange(0, max_len, dtype=torch.float).view(-1, 1)
        division_term = torch.exp(
            torch.arange(0, dim_model, 2).float() * (-np.log(10000.0)) / dim_model
        )
        pos_encoding[:, 0::2] = torch.sin(positions_list * division_term)
        pos_encoding[:, 1::2] = torch.cos(positions_list * division_term)

        pos_encoding = pos_encoding.unsqueeze(0).transpose(0, 1)
        self.register_buffer("pos_encoding", pos_encoding)

    def forward(self, token_embedding: torch.tensor) -> torch.tensor:
        return self.dropout(
            token_embedding + self.pos_encoding[: token_embedding.size(0), :]
        )


class AlignmentTransformer(nn.Module):
    """ """

    # Constructor
    def __init__(
        self,
        token_number=91,
        dim_model=128,
        dim_class=2,
        num_heads=4,
        num_decoder_layers=6,
        dropout_p=0.1,
    ):
        super().__init__()

        self.tokennumber = token_number
        self.dim_model = dim_model
        self.dim_class = dim_class
        self.num_heads = num_heads
        self.num_decoder_layers = num_decoder_layers
        # LAYERS
        self.positional_encoder = PositionalEncoding(
            dim_model=self.dim_model, dropout_p=dropout_p, max_len=50000
        )
        self.embedding = nn.Embedding(self.tokennumber, self.dim_model)

        # DECODER LAYERS
        D_layers = TransformerEncoderLayer(
            self.dim_model,
            nhead=self.num_heads,
            dim_feedforward=self.dim_model,
            dropout=dropout_p,
        )

        self.transformerDECODER = TransformerEncoder(
            encoder_layer=D_layers,
            num_layers=num_decoder_layers,
        )
        self.out = nn.Linear(self.dim_model, self.dim_class)

    def forward(self, src, tgt_mask=None, tgt_pad_mask=None):
        src = self.embedding(src)
        src = torch.sum(src, dim=-2)
        src = src.permute(1, 0, 2)
        src = self.positional_encoder(src)
        transformer_out = self.transformerDECODER(
            src=src, mask=tgt_mask, src_key_padding_mask=tgt_pad_mask
        )
        out = self.out(transformer_out)

        return out

    def get_tgt_mask(self, size) -> torch.tensor:
        mask = torch.zeros(size, size)
        return mask

    def create_pad_mask(
        self, matrix: torch.tensor, pad_token: int = -1
    ) -> torch.tensor:
        return matrix == pad_token


# TheGlueNote


class TheGlueNote(nn.Module):
    """
    GlueNote non-causal transformer encoder
    with learned positional encoding
    """

    # Constructor
    def __init__(
        self,
        device,
        token_number=314,
        position_number=512,
        dim_model=128,
        dim_feedforward=None,
        num_heads=8,
        num_decoder_layers=4,
        dropout_p=0.0,
        activation=nn.GELU(),
        using_decoder=True,
    ):
        super().__init__()
        self.device = device
        self.token_number = token_number
        self.position_number = position_number + 1
        self.dim_model = dim_model
        self.activation = activation
        if dim_feedforward is not None:
            self.dim_feedforward = dim_feedforward
        else:
            self.dim_feedforward = self.dim_model * 4
        self.num_heads = num_heads
        self.num_decoder_layers = num_decoder_layers
        self.dropout_p = dropout_p
        self.using_decoder = using_decoder
        # LAYERS
        self.positions = torch.arange(self.position_number * 2).to(self.device)
        self.positional_encoder = nn.Embedding(self.position_number * 2, self.dim_model)
        self.embedding = nn.Embedding(self.token_number, self.dim_model)

        # DECODER LAYERS
        D_layers = TransformerEncoderLayer(
            self.dim_model,
            nhead=self.num_heads,
            dim_feedforward=self.dim_feedforward,
            dropout=self.dropout_p,
            norm_first=True,
            activation=self.activation,
        )

        self.layer_normalization = LayerNorm(self.dim_model)
        self.layer_normalization_input = LayerNorm(self.dim_model)

        self.transformerDECODER = TransformerEncoder(
            encoder_layer=D_layers,
            num_layers=num_decoder_layers,
            norm=self.layer_normalization,
            enable_nested_tensor=False,
        )

        self.mlp1 = nn.Linear(self.dim_model, self.dim_feedforward)
        self.mlp_activation = activation = self.activation
        self.mlp2 = nn.Linear(self.dim_feedforward, self.dim_model)

        self.embed_out = nn.Linear(self.dim_model, self.dim_model)

    def forward(self, src, return_confidence_matrix=False):
        # # let's just only use pitch for now, no aggregation
        # src = src[:,1::4]
        src = self.embedding(src)

        # AGGREGATING 4 ATTRIBUTES
        eshape = src.shape
        src = torch.sum(src.reshape(eshape[0], -1, 4, eshape[-1]), dim=-2)

        # to obtain size (sequence length, batch_size, dim_model)
        src = src.permute(1, 0, 2)
        # POSITIONAL ENCODING
        pos = self.positional_encoder(self.positions)
        src += pos.unsqueeze(1)
        # src = self.positional_encoder(src)
        src = self.layer_normalization_input(src)
        # Transformer blocks - Out size = (sequence length, batch_size, dim_model)
        transformer_out = self.transformerDECODER(src=src)
        # mlp_out = self.mlp2(self.melp_activation(self.mlp1(transformer_out)))
        mlp_out = self.embed_out(transformer_out)
        predictions = torch.einsum(
            "ibk,jbk->bij",
            mlp_out[: self.position_number, :, :],
            mlp_out[self.position_number :, :, :],
        )  #

        if return_confidence_matrix:
            conf_matrix = F.softmax(predictions, 1) * F.softmax(predictions, 2)
            return conf_matrix
        else:
            if self.using_decoder:
                return predictions, mlp_out
            else:
                return predictions



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import io
import os
from setuptools import find_packages, setup

# Package meta-data.
NAME = "parangonar"
DESCRIPTION = "Symbolic music alignment"
KEYWORDS = "match alignment midi performance score"
URL = "https://github.com/sildater/parangonar"
AUTHOR = "Silvan Peter, Carlos Cancino-Chacón, Florian Henkel"
REQUIRES_PYTHON = ">=3.7"
VERSION = "2.0.0"

# What packages are required for this module to be executed?
REQUIRED = [
    "numpy",
    "scipy",
    "torch",
    "matplotlib",
    "partitura>=1.1.0",
    "symusic",
    "scipy",
    "miditok",
]

here = os.path.abspath(os.path.dirname(__file__))

try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

# Load the package's __version__.py module as a dictionary.
about = {}
if not VERSION:
    with open(os.path.join(here, NAME, "__version__.py")) as f:
        exec(f.read(), about)
else:
    about["__version__"] = VERSION

# Where the magic happens:
setup(
    name=NAME,
    version=about["__version__"],
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=KEYWORDS,
    author=AUTHOR,
    url=URL,
    python_requires=REQUIRES_PYTHON,
    packages=find_packages(exclude=("tests",)),
    package_data={
        "parangonar": [
            "assets/mozart_k265_var1.match",
        ]
    },
    install_requires=REQUIRED,
    extras_require={},
    include_package_data=True,
    license="Apache 2.0",
    classifiers=[
        # Trove classifiers
        # Full list: https://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
    ],
)



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains TempoOLTW.
"""

import numpy as np
import scipy
from enum import IntEnum
from queue import Queue
from scipy.spatial.distance import cdist
import matplotlib.pyplot as plt
import os

os.environ["KMP_DUPLICATE_LIB_OK"] = "True"


def tempo_and_pitch_metric(
    pitch_set_s,
    pitch_p,
    onset_s,
    onset_p,
    prev_onset_s,
    prev_onset_p,
    tempo,  # sec / beat
    time_weight=0.5,
    tempo_factor=0.5,
):
    """
    metric that combines
    1) pitch set metric
    2) onset deviation based on tempo

    returns the distance and the tempo associated with it
    """
    # pitch stuff
    if pitch_p in pitch_set_s:
        pitch_dist = 0
    else:
        pitch_dist = 1

    # onset stuff
    estimated_onset = prev_onset_p + (onset_s - prev_onset_s) * tempo
    onset_dist = abs(onset_p - estimated_onset) / tempo  # normalize offset by tempo4

    # tempo stuff
    if onset_s - prev_onset_s > 0:
        current_tempo = (onset_p - prev_onset_p) / (onset_s - prev_onset_s)
    else:
        current_tempo = tempo + (onset_p - prev_onset_p)
    exponential_average_tempo = (
        tempo_factor * current_tempo + (1 - tempo_factor) * tempo
    )

    dist = pitch_dist + time_weight * onset_dist
    return dist, exponential_average_tempo


def accumulate_tester():
    score = [[0, {1, 2}], [1, {3, 4}], [2, {3, 4}], [3, {3, 4}]]  # onset_s, pitch set
    perf = [
        [0, 1],
        [0.1, 2],
        [2, 3],
        [2.9, 4],
        [4, 3],
        [4.02, 4],
        [6, 4],
        [6.01, 3],
    ]

    M = len(perf)
    N = len(score)
    # the distance matrix is initialized with INFINITY
    D = np.ones((M + 1, N + 1), dtype=float) * np.inf
    D[0, 0] = 0
    # Tempo collector
    init_tempo = 2  # sec / beat
    Tempos = np.ones((M, N), dtype=float) * init_tempo
    directions = np.array([[1, 0], [1, 1], [0, 1]])

    for i in range(1, M + 1):
        for j in range(1, N + 1):
            print("+" * 50)
            mincost = np.inf
            besttempo = init_tempo
            for directionsidx, direction in enumerate(directions):
                (istep, jstep) = direction
                previ = i - istep
                prevj = j - jstep

                if previ > 0 and prevj > 0:
                    prev_onset_p = perf[previ - 1][0]
                    prev_onset_s = score[prevj - 1][0]
                    tempo = Tempos[previ - 1, prevj - 1]
                else:
                    prev_onset_p = 0  # smart?
                    prev_onset_s = 0  # smart?
                    tempo = init_tempo

                dist, tempo_new = tempo_and_pitch_metric(
                    pitch_set_s=score[j - 1][1],
                    pitch_p=perf[i - 1][1],
                    onset_s=score[j - 1][0],
                    onset_p=perf[i - 1][0],
                    prev_onset_s=prev_onset_s,
                    prev_onset_p=prev_onset_p,
                    tempo=tempo,  # sec / beat
                    time_weight=0.5,
                    tempo_factor=0.1,
                )
                print("*" * 20)
                print(i, j)
                print(previ, prevj)
                print("onsets s p", score[j - 1][0], perf[i - 1][0])
                print("pitches s p", score[j - 1][1], perf[i - 1][1])
                print(dist, tempo)

                if previ >= 0 and prevj >= 0:
                    cost = D[previ, prevj] + dist
                    if cost < mincost:
                        mincost = cost
                        besttempo = tempo_new

            D[i, j] = mincost
            Tempos[i - 1, j - 1] = besttempo

    return D, Tempos


class Direction(IntEnum):
    REF = 0
    INPUT = 1
    BOTH = 2

    def toggle(self):
        return Direction(self ^ 1) if self != Direction.BOTH else Direction.INPUT


class T_OLTW(object):
    """
    On-line Dynamic Time Warping (OLTW) algorithm for aligning a input sequence to a reference sequence.

    Adapted from: https://github.com/laurenceyoon/python-match

    Parameters
    ----------
    reference_features : list
        list of score features: pitch set and onset

    queue : Queue
        A queue for storing the input features, pitch and onset

    window_size : int
        Size of the window (in steps) of the cost matrix.

    max_run_count : int
        Maximum number of times the class can run in the same direction.

    hop_size : int
        number of seq items that get added at step

    directional_weights: np.ndarray
        weights associated with each of the three possible steps

    Attributes
    ----------
    warping_path : np.ndarray [shape=(2, T)]
        Warping path with pairs of indices of the reference and input features.
        where warping_path[0] is the index of the reference feature and
        warping_path[1] is the index of the input feature.
    """

    def __init__(
        self,
        reference_features=None,
        queue=None,
        window_size=10,  # shape of the acc cost matric
        max_run_count=100,  # maximal number of steps
        hop_size=1,  # number of seq items that get added at step
        directional_weights=np.array([1, 1, 1]),  # down, diag, right
        directions=np.array([[1, 0], [1, 1], [0, 1]]),
        metric=tempo_and_pitch_metric,
        init_tempo=1,  # sec / beat
        time_weight=0.5,
        tempo_factor=0.1,
        **kwargs,
    ):
        self.queue = queue
        self.metric = metric
        self.directional_weights = directional_weights
        self.directions = directions
        if reference_features is not None:
            self.set_feature_arrays(reference_features)
        else:
            self.reference_features = None
            self.N_ref = None
            self.input_features = None

        self.init_tempo = init_tempo
        self.time_weight = time_weight
        self.tempo_factor = tempo_factor
        self.w = window_size
        self.max_run_count = max_run_count
        self.hop_size = hop_size
        self.initialize()

    def set_feature_arrays(self, reference_features):
        self.reference_features = reference_features
        self.N_ref = len(reference_features)
        # prepend the first onset
        self.reference_features.insert(0, self.reference_features[0])
        self.input_features = list()

    def initialize(self):
        self.ref_pointer = 0
        self.input_pointer = 0
        self.run_count = 0
        self.previous_direction = None
        self.wp = np.array([[-1, -1]]).T  # [shape=(2, T)]

        self.ref_pointer += self.w  # window of ref shifted at start
        self.acc_dist_matrix = np.full((self.w, self.w), np.inf)
        self.acc_dist_matrix[0, :] = 0
        self.acc_len_matrix = np.zeros((self.w, self.w))
        self.queue_non_empty = True

    @property
    def warping_path(self):  # [shape=(2, T)]
        return self.wp[:, 1:]

    def offset(self):
        offset_x = max(self.ref_pointer - self.w, 0)
        offset_y = max(self.input_pointer - self.w, 0)
        return np.array([offset_x, offset_y])

    def update_ref_direction(
        self,
        new_tempo_acc,
        new_acc,
        new_len_acc,
        wx,
        wy,
        d,
        score_features,
        perf_features,
    ):
        update_x0 = wx - d
        for i in range(d):  # score direction
            for j in range(wy):  # performance direction
                mincost = np.inf
                besttempo = self.init_tempo
                bestpath = 1
                if j == 0:
                    prev_onset_p = perf_features[j + 1][
                        0
                    ]  # previous onset in p direction
                    prev_onset_s = score_features[i][0]
                    tempo = new_tempo_acc[update_x0 + i - 1, j]
                    local_dist, new_tempo = tempo_and_pitch_metric(
                        pitch_set_s=score_features[i + 1][1],
                        pitch_p=perf_features[j + 1][1],
                        onset_s=score_features[i + 1][0],
                        onset_p=perf_features[j + 1][0],
                        prev_onset_s=prev_onset_s,
                        prev_onset_p=prev_onset_p,
                        tempo=tempo,  # sec / beat
                        time_weight=self.time_weight,
                        tempo_factor=self.tempo_factor,
                    )

                    new_acc[update_x0 + i, j] = (
                        local_dist * self.directional_weights[0]
                        + new_acc[update_x0 + i - 1, j]
                    )
                    new_len_acc[update_x0 + i, j] = (
                        1 + new_len_acc[update_x0 + i - 1, j]
                    )
                    new_tempo_acc[update_x0 + i, j] = new_tempo

                else:
                    for d_idx, direction in enumerate(self.directions):
                        (istep, jstep) = direction
                        previ = i - istep
                        prevj = j - jstep
                        prev_onset_p = perf_features[prevj + 1][
                            0
                        ]  # previous onset in p direction
                        prev_onset_s = score_features[previ + 1][0]
                        tempo = new_tempo_acc[update_x0 + previ, prevj]
                        prev_path_len = new_len_acc[update_x0 + previ, prevj]
                        local_dist, new_tempo = tempo_and_pitch_metric(
                            pitch_set_s=score_features[i + 1][1],
                            pitch_p=perf_features[j + 1][1],
                            onset_s=score_features[i + 1][0],
                            onset_p=perf_features[j + 1][0],
                            prev_onset_s=prev_onset_s,
                            prev_onset_p=prev_onset_p,
                            tempo=tempo,  # sec / beat
                            time_weight=self.time_weight,
                            tempo_factor=self.tempo_factor,
                        )

                        cost = (
                            new_acc[update_x0 + previ, prevj]
                            + local_dist * self.directional_weights[d_idx]
                        )
                        if cost < mincost:
                            mincost = cost
                            besttempo = new_tempo
                            bestpath = 1 + prev_path_len

                    new_acc[update_x0 + i, j] = mincost
                    new_len_acc[update_x0 + i, j] = bestpath
                    new_tempo_acc[update_x0 + i, j] = besttempo

        return new_acc, new_len_acc, new_tempo_acc

    def update_input_direction(
        self,
        new_tempo_acc,
        new_acc,
        new_len_acc,
        wx,
        wy,
        d,
        score_features,
        perf_features,
    ):
        update_y0 = wy - d
        for i in range(wx):  # score direction
            for j in range(d):  # perfpormance direction
                mincost = np.inf
                besttempo = self.init_tempo
                bestpath = 1
                if i == 0:
                    prev_onset_p = perf_features[j][0]  # previous onset in p direction
                    prev_onset_s = score_features[i + 1][0]
                    tempo = new_tempo_acc[i, update_y0 + j - 1]
                    local_dist, new_tempo = tempo_and_pitch_metric(
                        pitch_set_s=score_features[i + 1][1],
                        pitch_p=perf_features[j + 1][1],
                        onset_s=score_features[i + 1][0],
                        onset_p=perf_features[j + 1][0],
                        prev_onset_s=prev_onset_s,
                        prev_onset_p=prev_onset_p,
                        tempo=tempo,  # sec / beat
                        time_weight=self.time_weight,
                        tempo_factor=self.tempo_factor,
                    )

                    new_acc[i, update_y0 + j] = (
                        local_dist * self.directional_weights[2]
                        + new_acc[i, update_y0 - 1 + j]
                    )
                    new_len_acc[i, update_y0 + j] = (
                        1 + new_len_acc[i, update_y0 - 1 + j]
                    )
                    new_tempo_acc[i, update_y0 + j] = new_tempo

                else:
                    for d_idx, direction in enumerate(self.directions):
                        (istep, jstep) = direction
                        previ = i - istep
                        prevj = j - jstep
                        prev_onset_p = perf_features[prevj + 1][
                            0
                        ]  # previous onset in p direction
                        prev_onset_s = score_features[previ + 1][0]
                        tempo = new_tempo_acc[previ, update_y0 + prevj]
                        prev_path_len = new_len_acc[previ, update_y0 + prevj]
                        local_dist, new_tempo = tempo_and_pitch_metric(
                            pitch_set_s=score_features[i + 1][1],
                            pitch_p=perf_features[j + 1][1],
                            onset_s=score_features[i + 1][0],
                            onset_p=perf_features[j + 1][0],
                            prev_onset_s=prev_onset_s,
                            prev_onset_p=prev_onset_p,
                            tempo=tempo,  # sec / beat
                            time_weight=self.time_weight,
                            tempo_factor=self.tempo_factor,
                        )
                        # print("*"*20)
                        # print(i, j)
                        # print(previ, prevj)
                        # print("onsets s p", score[j-1][0], perf[i-1][0])
                        # print("pitches s p", score[j-1][1], perf[i-1][1])
                        # print(dist, tempo)

                        cost = (
                            new_acc[previ, update_y0 + prevj]
                            + local_dist * self.directional_weights[d_idx]
                        )
                        if cost < mincost:
                            mincost = cost
                            besttempo = new_tempo
                            bestpath = 1 + prev_path_len

                    new_acc[i, update_y0 + j] = mincost
                    new_len_acc[i, update_y0 + j] = bestpath
                    new_tempo_acc[i, update_y0 + j] = besttempo

        return new_acc, new_len_acc, new_tempo_acc

    def update_both_direction(
        self,
        new_tempo_acc,
        new_acc,
        new_len_acc,
        wx,
        wy,
        d,
        score_features,
        perf_features,
    ):
        for i in range(wx):  # score direction
            for j in range(wy):  # perfpormance direction
                mincost = np.inf
                besttempo = self.init_tempo
                bestpath = 1
                if i == 0 and j >= wy - d:
                    prev_onset_p = perf_features[j][0]  # previous onset in p direction
                    prev_onset_s = score_features[i + 1][0]
                    tempo = new_tempo_acc[i, j - 1]
                    local_dist, new_tempo = tempo_and_pitch_metric(
                        pitch_set_s=score_features[i + 1][1],
                        pitch_p=perf_features[j + 1][1],
                        onset_s=score_features[i + 1][0],
                        onset_p=perf_features[j + 1][0],
                        prev_onset_s=prev_onset_s,
                        prev_onset_p=prev_onset_p,
                        tempo=tempo,  # sec / beat
                        time_weight=self.time_weight,
                        tempo_factor=self.tempo_factor,
                    )

                    new_acc[i, j] = (
                        local_dist * self.directional_weights[2] + new_acc[i, j - 1]
                    )
                    new_len_acc[i, j] = 1 + new_len_acc[i, j - 1]
                    new_tempo_acc[i, j] = new_tempo

                elif i >= wx - d and j == 0:
                    prev_onset_p = perf_features[j + 1][
                        0
                    ]  # previous onset in p direction
                    prev_onset_s = score_features[i][0]
                    tempo = new_tempo_acc[i - 1, j]
                    local_dist, new_tempo = tempo_and_pitch_metric(
                        pitch_set_s=score_features[i + 1][1],
                        pitch_p=perf_features[j + 1][1],
                        onset_s=score_features[i + 1][0],
                        onset_p=perf_features[j + 1][0],
                        prev_onset_s=prev_onset_s,
                        prev_onset_p=prev_onset_p,
                        tempo=tempo,  # sec / beat
                        time_weight=self.time_weight,
                        tempo_factor=self.tempo_factor,
                    )

                    new_acc[i, j] = (
                        local_dist * self.directional_weights[0] + new_acc[i - 1, j]
                    )
                    new_len_acc[i, j] = 1 + new_len_acc[i - 1, j]
                    new_tempo_acc[i, j] = new_tempo

                elif i >= wx - d or j >= wy - d:
                    for d_idx, direction in enumerate(self.directions):
                        (istep, jstep) = direction
                        previ = i - istep
                        prevj = j - jstep
                        prev_onset_p = perf_features[prevj + 1][
                            0
                        ]  # previous onset in p direction
                        prev_onset_s = score_features[previ + 1][0]
                        tempo = new_tempo_acc[previ, prevj]
                        prev_path_len = new_len_acc[previ, prevj]
                        local_dist, new_tempo = tempo_and_pitch_metric(
                            pitch_set_s=score_features[i + 1][1],
                            pitch_p=perf_features[j + 1][1],
                            onset_s=score_features[i + 1][0],
                            onset_p=perf_features[j + 1][0],
                            prev_onset_s=prev_onset_s,
                            prev_onset_p=prev_onset_p,
                            tempo=tempo,  # sec / beat
                            time_weight=self.time_weight,
                            tempo_factor=self.tempo_factor,
                        )

                        cost = (
                            new_acc[previ, prevj]
                            + local_dist * self.directional_weights[d_idx]
                        )
                        if cost < mincost:
                            mincost = cost
                            besttempo = new_tempo
                            bestpath = 1 + prev_path_len

                    new_acc[i, j] = mincost
                    new_len_acc[i, j] = bestpath
                    new_tempo_acc[i, j] = besttempo

        return new_acc, new_len_acc, new_tempo_acc

    def update_cost_matrix(self, direction):
        # local cost matrix
        x, y = self.ref_pointer, self.input_pointer
        wx, wy = min(self.w, x), min(self.w, y)
        d = self.hop_size
        # storage matrices
        new_acc = np.full((wx, wy), np.inf, dtype=np.float32)
        new_len_acc = np.zeros((wx, wy))
        new_tempo_acc = np.full((wx, wy), self.init_tempo, dtype=float)

        if direction is Direction.REF:
            new_acc[:-d, :] = self.acc_dist_matrix[d:]
            new_len_acc[:-d, :] = self.acc_len_matrix[d:]
            new_tempo_acc[:-d, :] = self.acc_tempo_matrix[d:, :]
            x_seg = self.reference_features[x - d : x + 1]  # [d + 1]
            y_seg = self.input_features[y - wy : y + 1]  # [wy + 1]
            # x_seg = self.reference_features[x - d : x]  # [d, 12]
            # y_seg = self.input_features[y - wy : y]  # [wy, 12]

            new_acc, new_len_acc, new_tempo_acc = self.update_ref_direction(
                new_tempo_acc, new_acc, new_len_acc, wx, wy, d, x_seg, y_seg
            )

        elif direction is Direction.INPUT:
            overlap_y = wy - d
            new_acc[:, :-d] = self.acc_dist_matrix[:, -overlap_y:]
            new_len_acc[:, :-d] = self.acc_len_matrix[:, -overlap_y:]
            new_tempo_acc[:, :-d] = self.acc_tempo_matrix[:, -overlap_y:]
            # x_seg = self.reference_features[x - wx : x]  # [wx, 12]
            # y_seg = self.input_features[y - d : y]  # [d, 12]
            x_seg = self.reference_features[x - wx : x + 1]  # [wx + 1]
            y_seg = self.input_features[y - d : y + 1]  # [d + 1]
            new_acc, new_len_acc, new_tempo_acc = self.update_input_direction(
                new_tempo_acc, new_acc, new_len_acc, wx, wy, d, x_seg, y_seg
            )

        elif direction is Direction.BOTH:  # input, BOTH
            overlap_y = wy - d
            new_acc[:-d, :-d] = self.acc_dist_matrix[d:, -overlap_y:]
            new_len_acc[:-d, :-d] = self.acc_len_matrix[d:, -overlap_y:]
            new_tempo_acc[:-d, :-d] = self.acc_tempo_matrix[d:, -overlap_y:]
            # x_seg = self.reference_features[x - wx : x]  # [wx, 12]
            # y_seg = self.input_features[y - wy : y]  # [wy, 12]
            x_seg = self.reference_features[x - wx : x + 1]  # [wx + 1]
            y_seg = self.input_features[y - wy : y + 1]  # [wy  + 1]
            new_acc, new_len_acc, new_tempo_acc = self.update_both_direction(
                new_tempo_acc, new_acc, new_len_acc, wx, wy, d, x_seg, y_seg
            )

        self.acc_dist_matrix = new_acc
        self.acc_len_matrix = new_len_acc
        self.acc_tempo_matrix = new_tempo_acc

    def first_cost_matrix(self):
        # local cost matrix
        x, y = self.ref_pointer, self.input_pointer  # should be window, hop_size
        wx, wy = x, y + 1  # append an extra column
        d = self.hop_size
        # storage matrices
        new_acc = np.full((wx, wy), np.inf, dtype=np.float32)
        new_acc[0, 0] = 0  # starting point in corner
        new_len_acc = np.zeros((wx, wy))
        new_tempo_acc = np.full((wx, wy), self.init_tempo, dtype=float)

        # copy the first element and extract overlapping feature vectors
        self.input_features.insert(0, self.input_features[0])
        x_seg = self.reference_features[x - wx : x + 1]  # [wx + 1]
        y_seg = self.input_features[y - d : y + 1]  # [d + 1]
        # print("score at start:", x_seg)
        # print("perf at start:", y_seg)

        new_acc, new_len_acc, new_tempo_acc = self.update_input_direction(
            new_tempo_acc, new_acc, new_len_acc, wx, wy, d, x_seg, y_seg
        )

        self.acc_dist_matrix = new_acc[:, 1:]
        self.acc_len_matrix = new_len_acc[:, 1:]
        self.acc_tempo_matrix = new_tempo_acc[:, 1:]

    def select_candidate(self):
        norm_x_edge = self.acc_dist_matrix[-1, :] / self.acc_len_matrix[-1, :]
        norm_y_edge = self.acc_dist_matrix[:, -1] / self.acc_len_matrix[:, -1]
        cat = np.concatenate((norm_x_edge, norm_y_edge))

        min_idx = np.argmin(cat)
        offset = self.offset()
        if min_idx < len(norm_x_edge):
            self.candidate = np.array([self.ref_pointer - offset[0] - 1, min_idx])
        else:
            self.candidate = np.array(
                [min_idx - len(norm_x_edge), self.input_pointer - offset[1] - 1]
            )

        # print("edge",cat)
        # print("pointers:", self.ref_pointer, self.input_pointer, offset)
        # print("cand",self.candidate, min_idx)

    def add_candidate_to_path(self):
        new_coordinate = np.expand_dims(
            self.offset() + self.candidate, axis=1
        )  # [2, 1]
        self.wp = np.concatenate((self.wp, new_coordinate), axis=1)

    def select_next_direction(self):
        if self.input_pointer < self.w:
            next_direction = Direction.INPUT
        elif self.run_count > self.max_run_count:
            next_direction = self.previous_direction.toggle()
        else:
            offset = self.offset()
            x0, y0 = offset[0], offset[1]
            ref_bool = self.candidate[0] == self.ref_pointer - x0 - 1
            input_bool = self.candidate[1] == self.input_pointer - y0 - 1
            if ref_bool and input_bool:
                next_direction = Direction.BOTH
            elif ref_bool:
                next_direction = Direction.REF
            else:
                assert input_bool
                next_direction = Direction.INPUT
        return next_direction

    def get_new_input(self):
        try:
            input_feature = self.queue.get(block=False)
            self.input_features += input_feature
            self.input_pointer += self.hop_size

        except:
            print("empty queue")
            self.queue_non_empty = False

    def is_still_following(self):
        is_still_following = self.ref_pointer <= (self.N_ref - self.hop_size)
        return is_still_following and self.queue_non_empty

    def handle_direction(self, direction):
        # get fitting features/windows
        if direction is not Direction.REF:
            self.get_new_input()
        if direction is not Direction.INPUT:
            self.ref_pointer += self.hop_size

        # accumulate run count
        if direction == self.previous_direction:
            self.run_count += 1
        else:
            self.run_count = 1
        self.previous_direction = direction

    def handle_first_input(self):
        direction = self.select_next_direction()
        self.handle_direction(direction)
        self.first_cost_matrix()
        self.select_candidate()
        self.add_candidate_to_path()

    def run(self, verbose=False):
        print("Start running OLTW")
        self.initialize()
        self.handle_first_input()
        while self.is_still_following():
            direction = self.select_next_direction()
            if verbose:
                print("ACC DIST \n", self.acc_dist_matrix)
                print("ACC LEN \n", self.acc_len_matrix)
                print("TEMPO \n", self.acc_tempo_matrix)
                print(
                    "RECENT WARPING PATH (top s, bottom p) \n",
                    self.warping_path[:, -3:],
                )
                print("NEXT DIRECTION \n", direction)
                print("*" * 50)
                # plt.imshow(self.acc_dist_matrix)
                # plt.savefig("score_{0}_perf_{1}.png".format(self.ref_pointer, self.input_pointer))
                # plt.close()
            self.handle_direction(direction)
            self.update_cost_matrix(direction)
            self.select_candidate()
            self.add_candidate_to_path()
        print("... and we're done.")
        return self.warping_path


def testfeatures_t_oltw():
    score = [
        [0, {1, 2}],  # onset_s, pitch set
        [1, {3, 4}],
        [2, {3, 4}],
        [3, {3, 4}],
        [4, {3, 4}],
        [5, {3, 4}],
        [6, {3, 4}],
        [7, {3, 4}],
    ]
    perf = [
        [0, 1],
        [0.1, 2],
        [2, 3],
        [2.1, 4],
        [4, 3],
        [4.02, 4],
        [6, 4],
        [6.01, 3],
    ]
    return score, perf


if __name__ == "__main__":
    RANGE_L = 6
    REPEATS = 3
    HOP_SIZE = 1
    WINDOW_SIZE = 3

    r, t = testfeatures_t_oltw()
    queue = Queue()

    for tt in t:
        queue.put([tt])

    o = T_OLTW(
        reference_features=r,
        queue=queue,
        frame_per_seg=HOP_SIZE,
        window_size=WINDOW_SIZE,
        max_run_count=4,
        init_tempo=2,
        tempo_factor=0.1,
        time_weight=0.1,
        directional_weights=np.array([1.0, 1.0, 1.0]),
    )
    # p = o.run()
    # print("path \n", p)

    # D, T = accumulate_tester()



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains methods to visualize and
evaluate note alignments of symbolic music data.
"""
from typing import List
import numpy as np
import partitura as pt
import matplotlib.pyplot as plt
import random
import os


def fscore_alignments(
    prediction: List[dict],
    ground_truth: List[dict],
    types: List[str],
    return_numbers=False,
) -> (float, float, float):
    """
    Parameters
    ----------
    prediction: List of dictionaries containing the predicted alignments
    ground_truth: List of dictionaries containing the ground truth alignments
    types: List of alignment types to consider for evaluation (e.g ['match', 'deletion', 'insertion']

    Returns
    -------
    precision, recall, f score
    """

    pred_filtered = list(filter(lambda x: x["label"] in types, prediction))
    gt_filtered = list(filter(lambda x: x["label"] in types, ground_truth))

    filtered_correct = [pred for pred in pred_filtered if pred in gt_filtered]

    n_pred_filtered = len(pred_filtered)
    n_gt_filtered = len(gt_filtered)
    n_correct = len(filtered_correct)

    if n_pred_filtered > 0 or n_gt_filtered > 0:
        precision = n_correct / n_pred_filtered if n_pred_filtered > 0 else 0.0
        recall = n_correct / n_gt_filtered if n_gt_filtered > 0 else 0
        f_score = (
            2 * precision * recall / (precision + recall)
            if (precision + recall) > 0
            else 0
        )
    else:
        # no prediction and no ground truth for a given type -> correct alignment
        precision, recall, f_score = 1.0, 1.0, 1.0

    if return_numbers:
        return precision, recall, f_score, len(pred_filtered), len(gt_filtered)
    else:
        return precision, recall, f_score


def print_fscore_alignments(prediction: List[dict], ground_truth: List[dict]):
    print("------------------")
    types = ["match", "insertion", "deletion"]
    for alignment_type in types:
        precision, recall, f_score = fscore_alignments(
            prediction, ground_truth, alignment_type
        )
        print("Evaluate ", alignment_type)
        print(
            "Precision: ",
            format(precision, ".3f"),
            "Recall ",
            format(recall, ".3f"),
            "F-Score ",
            format(f_score, ".3f"),
        )
        print("------------------")


def plot_alignment(
    ppart_na,
    part_na,
    alignment,
    save_file=False,
    fname="note_alignment",
    random_color=False,
):
    first_note_midi = np.min(ppart_na["onset_sec"])
    last_note_midi = np.max(ppart_na["onset_sec"] + ppart_na["duration_sec"])
    first_note_start = np.min(part_na["onset_beat"])
    last_note_start = np.max(part_na["onset_beat"])
    length_of_midi = last_note_midi - first_note_midi
    length_of_xml = last_note_start - first_note_start

    length_of_pianorolls = max(10000, int(length_of_xml * 8))
    time_div_midi = int(np.floor(length_of_pianorolls / length_of_midi))
    time_div_xml = int(np.floor(length_of_pianorolls / length_of_xml))

    midi_piano_roll, perfidx = pt.utils.compute_pianoroll(
        ppart_na,
        time_unit="sec",
        time_div=time_div_midi,
        return_idxs=True,
        remove_drums=False,
    )
    xml_piano_roll, scoreidx = pt.utils.compute_pianoroll(
        part_na,
        time_unit="beat",
        time_div=time_div_xml,
        return_idxs=True,
        remove_drums=False,
    )

    plot_array = np.zeros((128 * 2 + 50, length_of_pianorolls + 800))
    dense_midi_pr = midi_piano_roll.todense()  # [:,:time_size]
    dense_midi_pr[dense_midi_pr > 0] = 1
    plot_array[:128, : xml_piano_roll.shape[1]] = xml_piano_roll.todense()
    plot_array[50 + 128 :, : midi_piano_roll.shape[1]] = dense_midi_pr

    f, axs = plt.subplots(1, 1, figsize=(100, 10))
    axs.matshow(plot_array, aspect="auto", origin="lower")
    hexadecimal_alphabets = "0123456789ABCDEF"

    if random_color:
        colors = [
            "#" + "".join([random.choice(hexadecimal_alphabets) for j in range(6)])
            for i in range(40)
        ]
    else:
        colors = ["#00FF00" for i in range(40)]

    score_dict = dict()
    for note, pos in zip(part_na, scoreidx):
        score_dict[note["id"]] = pos
    perf_dict = dict()
    for note, pos in zip(ppart_na, perfidx):
        perf_dict[note["id"]] = pos
    for i, line in enumerate(alignment):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors[i % 40],
            )

    if save_file:
        plt.savefig(fname + ".png")
        plt.close(f)
    else:
        plt.show()


def plot_alignment_comparison(
    ppart_na,
    part_na,
    alignment1,
    alignment2,
    save_file=False,
    return_figure=False,
    figsize=(100, 10),
    fname="note_alignments",
):
    first_note_midi = np.min(ppart_na["onset_sec"])
    last_note_midi = np.max(ppart_na["onset_sec"] + ppart_na["duration_sec"])
    first_note_start = np.min(part_na["onset_beat"])
    last_note_start = np.max(part_na["onset_beat"])
    length_of_midi = last_note_midi - first_note_midi
    length_of_xml = last_note_start - first_note_start

    length_of_pianorolls = max(10000, int(length_of_xml * 8))
    time_div_midi = int(np.floor(length_of_pianorolls / length_of_midi))
    time_div_xml = int(np.floor(length_of_pianorolls / length_of_xml))

    midi_piano_roll, perfidx = pt.utils.compute_pianoroll(
        ppart_na,
        time_unit="sec",
        time_div=time_div_midi,
        return_idxs=True,
        remove_drums=False,
    )
    xml_piano_roll, scoreidx = pt.utils.compute_pianoroll(
        part_na,
        time_unit="beat",
        time_div=time_div_xml,
        return_idxs=True,
        remove_drums=False,
    )

    plot_array = np.zeros((128 * 2 + 50, length_of_pianorolls + 800))
    dense_midi_pr = midi_piano_roll.todense()  # [:,:time_size]
    dense_midi_pr[dense_midi_pr > 0] = 1
    plot_array[:128, : xml_piano_roll.shape[1]] = xml_piano_roll.todense()
    plot_array[50 + 128 :, : midi_piano_roll.shape[1]] = dense_midi_pr

    f, axs = plt.subplots(1, 1, figsize=figsize)
    axs.matshow(plot_array, aspect="auto", origin="lower")
    hexadecimal_alphabets = "0123456789ABCDEF"

    colors1 = ["#00FF00" for i in range(40)]
    colors2 = ["#0000FF" for i in range(40)]
    colors3 = ["#FF0000" for i in range(40)]
    colors4 = ["#FF00FF" for i in range(40)]

    n1_but_not_n2 = [al for al in alignment1 if not al in alignment2]
    n2_but_not_n1 = [al for al in alignment2 if not al in alignment1]

    score_dict = dict()
    for note, pos in zip(part_na, scoreidx):
        score_dict[note["id"]] = pos
    perf_dict = dict()
    for note, pos in zip(ppart_na, perfidx):
        perf_dict[note["id"]] = pos
    for i, line in enumerate(alignment1):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors1[i % 40],
            )
    for i, line in enumerate(alignment2):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors2[i % 40],
            )
    for i, line in enumerate(n1_but_not_n2):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors3[i % 40],
            )
    for i, line in enumerate(n2_but_not_n1):
        if line["label"] == "match":
            perf_pos = perf_dict[line["performance_id"]]
            score_pos = score_dict[line["score_id"]]
            axs.plot(
                [score_pos[1], perf_pos[1]],
                [score_pos[0], 128 + 50 + perf_pos[0]],
                "o-",
                lw=2,
                c=colors4[i % 40],
            )

    if save_file:
        plt.savefig(fname + ".png")
        plt.close(f)
    else:
        if return_figure:
            return f
        else:
            plt.show()


def plot_alignment_mappings(
    ppart_na,
    part_na,
    score_to_performance_mapping1,
    score_to_performance_mapping2,
    save_file=False,
    fname="onset_alignments",
):
    first_note_midi = np.min(ppart_na["onset_sec"])
    last_note_midi = np.max(ppart_na["onset_sec"] + ppart_na["duration_sec"])
    first_note_start = np.min(part_na["onset_beat"])
    last_note_start = np.max(part_na["onset_beat"])
    length_of_midi = last_note_midi - first_note_midi
    length_of_xml = last_note_start - first_note_start

    length_of_pianorolls = max(10000, int(length_of_xml * 8))
    time_div_midi = int(np.floor(length_of_pianorolls / length_of_midi))
    time_div_xml = int(np.floor(length_of_pianorolls / length_of_xml))

    midi_piano_roll, perfidx = pt.utils.compute_pianoroll(
        ppart_na,
        time_unit="sec",
        time_div=time_div_midi,
        return_idxs=True,
        remove_drums=False,
    )
    xml_piano_roll, scoreidx = pt.utils.compute_pianoroll(
        part_na,
        time_unit="beat",
        time_div=time_div_xml,
        return_idxs=True,
        remove_drums=False,
    )

    midi_to_pr_mapping = lambda x: int((x - first_note_midi) * time_div_midi)
    xml_to_pr_mapping = lambda x: int((x - first_note_start) * time_div_xml)

    plot_array = np.zeros((128 * 2 + 50, length_of_pianorolls + 800))
    dense_midi_pr = midi_piano_roll.todense()  # [:,:time_size]
    dense_midi_pr[dense_midi_pr > 0] = 1
    plot_array[:128, : xml_piano_roll.shape[1]] = xml_piano_roll.todense()
    plot_array[50 + 128 :, : midi_piano_roll.shape[1]] = dense_midi_pr

    f, axs = plt.subplots(1, 1, figsize=(100, 10))
    axs.matshow(plot_array, aspect="auto", origin="lower")
    for unique_onset in np.unique(part_na["onset_beat"]):
        perf_pos1 = midi_to_pr_mapping(score_to_performance_mapping1(unique_onset))
        perf_pos2 = midi_to_pr_mapping(score_to_performance_mapping2(unique_onset))
        score_pos = xml_to_pr_mapping(unique_onset)
        axs.plot([score_pos, perf_pos1], [128, 128 + 50], "o-", lw=2, c="#00FF00")
        axs.plot([score_pos, perf_pos2], [128, 128 + 50], "o-", lw=2, c="#0000FF")

    if save_file:
        plt.savefig(fname + ".png")
        plt.close(f)
    else:
        plt.show()


def evaluate_asynchrony(target_ponsets, tracked_ponsets):
    asynchrony = target_ponsets - tracked_ponsets
    abs_asynch = abs(asynchrony)
    mean_asynch = np.median(abs_asynch)
    lt_25ms = np.mean(abs_asynch <= 0.025)
    lt_50ms = np.mean(abs_asynch <= 0.05)
    lt_100ms = np.mean(abs_asynch <= 0.1)
    return mean_asynch, lt_25ms, lt_50ms, lt_100ms


def evaluate_score_following(
    performance_note_array=None,
    score_note_array=None,
    gt_alignment=None,
    alignment=None,
    out_dir="",
    file_suffix="",
    write_to_file=False,
    print_results=False,
):
    """
    Parameters
    ----------
    performance_note_array: np.ndarray

    score_note_array: np.ndarray

    gt_alignment: List[dict]

    alignment: List[dict]

    out_dir: str

    print_results: bool

    Returns
    -------
    mean_asynch: float

    lt_25ms: float

    lt_50ms: float

    lt_100ms: float

    """

    (
        _,
        stime_to_ptime_map_gt,
    ) = pt.musicanalysis.performance_codec.get_time_maps_from_alignment(
        performance_note_array, score_note_array, gt_alignment, remove_ornaments=True
    )

    (
        _,
        stime_to_ptime_map,
    ) = pt.musicanalysis.performance_codec.get_time_maps_from_alignment(
        performance_note_array, score_note_array, alignment, remove_ornaments=True
    )
    sid1 = set()
    sid2 = set()
    for al in alignment:
        # Get only matched notes (i.e., ignore inserted or deleted notes)
        if al["label"] == "match":
            sid1.add(al["score_id"])
    for al in gt_alignment:
        if al["label"] == "match":
            sid2.add(al["score_id"])

    all_sid = list(sid1.intersection(sid2))

    tracked_sonsets = np.array(
        [
            score_note["onset_beat"]
            for score_note in score_note_array
            if score_note["id"] in all_sid
        ]
    )
    tracked_sonsets = np.unique(tracked_sonsets)

    target_ponsets = stime_to_ptime_map_gt(tracked_sonsets)
    tracked_ponsets = stime_to_ptime_map(tracked_sonsets)

    mean_asynch, lt_25ms, lt_50ms, lt_100ms = evaluate_asynchrony(
        target_ponsets=target_ponsets,
        tracked_ponsets=tracked_ponsets,
    )

    if write_to_file:
        results_fn = os.path.join(out_dir, file_suffix + "results.csv")
        with open(results_fn, "w") as f:
            f.write("mean_asynch_ms,leq_25ms_%,leq_50ms_%,leq_100ms_%\n")

            if print_results:
                print(
                    f"Mean asynchrony (ms): {mean_asynch * 1000:.2f}\nAsynchrony <= 25ms (%): {lt_25ms * 100:.1f}\n"
                    f"Asynchrony <= 50ms (%): {lt_50ms * 100:.1f}\nAsynchrony <= 100ms (%): {lt_100ms * 100:.1f}\n"
                )
            f.write(
                f"{mean_asynch * 1000:.2f},{lt_25ms * 100:.1f},"
                f"{lt_50ms * 100:.1f},{lt_100ms * 100:.1f}"
            )

    else:
        return mean_asynch, lt_25ms, lt_50ms, lt_100ms



#!/usr/bin/python
# -*- coding: utf-8 -*-
"""
This module contains dynamic time warping methods.
"""

import numpy as np
from scipy.spatial.distance import cdist
import numba
from numba import jit

# helpers and metrics


def element_of_metric(vec1, vec2):
    """
    metric that evaluates occurence of vec2 (scalar) in vec1 (vector n-dim)
    """
    return 1 - np.sum(vec2 == vec1)


def element_of_set_metric(element_, set_):
    """
    metric that evaluates occurence of vec2 (scalar) in vec1 (vector n-dim)
    """
    if element_ in set_:
        return 0.0
    else:
        return 1.0


def l2(vec1, vec2):
    """
    l2 metric between vec1 and vec2
    """
    return np.sqrt(np.sum((vec2 - vec1) ** 2))


def invert_matrix(S, inversion="reciprocal", positive=False):
    """
    simple converter from similarity to distance matrix
    and vice versa
    """
    if inversion == "reciprocal":
        D = 1 / (S + 1e-4)
    else:
        D = -S
    if positive:
        D -= D.min()
    return D


def dnw(vec1, vec2):
    """
    normalized and weighted distance for LNCO/LNSO features.
    https://ieeexplore.ieee.org/abstract/document/6333860/
    """
    vec1_l1 = np.abs(vec1).sum()
    vec2_l1 = np.abs(vec2).sum()
    dn = np.abs(vec1 - vec2).sum() / (vec1_l1 + vec2_l1 + 1e-7)  # L1 okay?
    dampening_factor = ((vec1_l1 + vec2_l1 + 1e-7) / 2) ** (
        1 / 4
    )  # safety eps not necessary
    return dn * dampening_factor


def cdist_local(arr1, arr2, metric):
    """
    compute array of pairwise distances between
    the elements of two arrays given a metric

    Parameters
    ----------
    arr1: numpy nd array

    arr2: numpy nd array

    metric: callable
        a metric function

    Returns
    -------

    pdist_array: numpy 2d array
        array of pairwise distances
    """
    pdist_array = np.ones((arr1.shape[0], arr2.shape[0])) * np.inf
    for i in range(arr1.shape[0]):
        for j in range(arr2.shape[0]):
            pdist_array[i, j] = metric(arr1[i], arr2[j])
    return pdist_array


# DTW classes


class WeightedDynamicTimeWarping(object):
    """
    Generalized Weighted Dynamic Time Warping.

    Parameters
    ----------
    directional_weights: np.ndarray
        weights associated with each of the three possible steps
    directions : double array
        directions.

    """

    def __init__(
        self,
        directional_weights=np.array([1, 1, 1]),
        directions=np.array([[1, 0], [1, 1], [0, 1]]),
        metric="euclidean",
        cdist_local=False,
    ):
        self.directional_weights = directional_weights
        self.directions = directions
        self.metric = metric
        self.cdist_local = cdist_local

    def __call__(self, X, Y, return_matrices=False, return_cost=False):
        """
        Parameters
        ----------
        X : np.ndarray
            sequence 1 features, 1 row per step.
        Y : np.ndarray
            sequence 2 features, 1 row per step.
        return_matrices: bool
            return accumulated cost matrix
        return_cost : bool
            return accumulated cost of the minimizing path.

        Returns
        -------
        path : np.ndarray
            Accumulated cost matrix
        """

        X = np.asanyarray(X, dtype=float)
        Y = np.asanyarray(Y, dtype=float)
        # Compute pairwise distance
        if self.cdist_local:
            pwD = cdist_local(X, Y, self.metric)
        else:
            pwD = cdist(X, Y, self.metric)

        out = self.from_distance_matrix(
            pwD, return_matrices=return_matrices, return_cost=return_cost
        )
        return out

    def from_distance_matrix(self, pwD, return_matrices=False, return_cost=False):
        """
            Parameters
        ----------
        pwD : np.ndarray
            pairwise distance matrix
        return_matrices: bool
            return accumulated costmatrix, backtracking, and
            starting point matrix
        return_cost : bool
            return accumulated cost of the minimizing path.

        Returns
        -------
        path : np.ndarray
            Accumulated cost matrix
        """

        D, path = weighted_dtw_forward_and_backward(
            pwD, self.directional_weights, self.directions
        )
        out = (path,)
        if return_matrices:
            out += (D,)
        if return_cost:
            out += (D[path[-1, 0], path[-1, 1]],)
        return out


# alias
WDTW = WeightedDynamicTimeWarping


class DynamicTimeWarping(object):
    """
    pure python vanilla Dynamic Time Warping
    """

    def __init__(self, metric="euclidean", cdist_local=False):
        self.metric = metric
        self.cdist_local = cdist_local

    def __call__(self, X, Y, return_path=True, return_cost_matrix=False):
        X = np.asanyarray(X, dtype=float)
        Y = np.asanyarray(Y, dtype=float)
        # Compute pairwise distance
        if self.cdist_local:
            D = cdist_local(X, Y, self.metric)
        else:
            D = cdist(X, Y, self.metric)
        # Compute accumulated cost matrix
        dtwd_matrix = dtw_dmatrix_from_pairwise_dmatrix(D)
        dtwd_distance = dtwd_matrix[-1, -1]

        # Output
        out = (dtwd_distance,)

        if return_path:
            # Compute alignment path
            path = dtw_backtracking(dtwd_matrix)
            out += (path,)
        if return_cost_matrix:
            out += (dtwd_matrix,)
        return out


# alias
DTW = DynamicTimeWarping


class DynamicTimeWarpingSingleLoop(object):
    """
    pure python vanilla Dynamic Time Warping
    """

    def __init__(self, metric=element_of_set_metric):
        self.metric = metric

    def __call__(self, X, Y, return_path=True, return_cost_matrix=False):
        # Compute the pw distances and accumulated cost matrix
        dtwd_matrix = cdist_dtw_single_loop(X, Y, self.metric)
        # dtwd_matrix = dtw_dmatrix_from_pairwise_dmatrix(D)
        dtwd_distance = dtwd_matrix[-1, -1]

        # Output
        out = (dtwd_distance,)

        if return_path:
            # Compute alignment path
            path = dtw_backtracking(dtwd_matrix)
            out += (path,)
        if return_cost_matrix:
            out += (dtwd_matrix,)
        return out


# alias
DTWSL = DynamicTimeWarpingSingleLoop


class FlexDynamicTimeWarping(object):
    """
    FlexDTW: https://ismir2023program.ismir.net/poster_235.html
    from two vectors

    Parameters
    ----------
    directional_weights: np.ndarray
        weights associated with each of the three possible steps
    directions : double array
        directions.
    buffer: int
        buffer zone for flexible path end point

    """

    def __init__(
        self,
        directional_weights=np.array([1, 1, 1]),
        directions=np.array([[1, 0], [1, 1], [0, 1]]),
        buffer=1,
        metric="euclidean",
        cdist_local=False,
    ):
        self.directional_weights = directional_weights
        self.directions = directions
        self.buffer = buffer
        self.metric = metric
        self.cdist_local = cdist_local

    def __call__(self, X, Y, return_matrices=False, return_cost=False):
        """
        Parameters
        ----------
        X : np.ndarray
            sequence 1 features, 1 row per step.
        Y : np.ndarray
            sequence 2 features, 1 row per step.
        return_matrices: bool
            return accumulated costmatrix, backtracking, and
            starting point matrix
        return_cost : bool
            return accumulated cost of the minimizing path.

        Returns
        -------
        path : np.ndarray
            Accumulated cost matrix
        """

        X = np.asanyarray(X, dtype=float)
        Y = np.asanyarray(Y, dtype=float)
        # Compute pairwise distance
        if self.cdist_local:
            pwD = cdist_local(X, Y, self.metric)
        else:
            pwD = cdist(X, Y, self.metric)

        out = self.from_distance_matrix(
            pwD, return_matrices=return_matrices, return_cost=return_cost
        )
        return out

    def from_distance_matrix(self, pwD, return_matrices=False, return_cost=False):
        """
            Parameters
        ----------
        pwD : np.ndarray
            pairwise distance matrix
        return_matrices: bool
            return accumulated costmatrix, backtracking, and
            starting point matrix
        return_cost : bool
            return accumulated cost of the minimizing path.

        Returns
        -------
        path : np.ndarray
            Accumulated cost matrix
        """
        path, D, B, S = flexdtw_forward_and_backward(
            pwD, self.directional_weights, self.directions, self.buffer
        )
        out = (path,)
        if return_matrices:
            out += (
                D,
                B,
                S,
            )
        if return_cost:
            out += D[path[-1, 0], path[-1, 1]]
        return out


# alias
FDTW = FlexDynamicTimeWarping

# DTW fw + bw


@jit(nopython=True)
def weighted_dtw_forward_and_backward(
    pwD,
    directional_weights=np.array([1, 1, 1]),
    directions=np.array([[1, 0], [1, 1], [0, 1]]),
):
    """
    compute dynamic time warping cost matrix
    and backtracking path
    from weighted directions and
    a pairwise distance matrix

    Parameters
    ----------
    D : np.ndarray
        Pairwise distance matrix (computed e.g., with `cdist`).
    directional_weights: np.ndarray
        weights associated with each of the three possible steps
    directions : double array
        directions.

    Returns
    -------
    dtwd : np.ndarray
        Accumulated cost matrix
    path: np.ndarray
        backtracked path
    """
    # Initialize arrays and helper variables
    M = pwD.shape[0]
    N = pwD.shape[1]
    # the dtwd distance matrix is initialized with INFINITY
    D = np.ones((M + 1, N + 1), dtype=float) * np.inf
    # Backtracking
    B = np.ones((M, N), dtype=np.int8) * -1

    # Compute the distance iteratively
    D[0, 0] = 0
    for i in range(1, M + 1):
        for j in range(1, N + 1):
            mincost = np.inf
            minidx = -1
            bestiprev = -1
            bestjprev = -1
            for directionsidx, direction in enumerate(directions):
                (istep, jstep) = direction
                previ = i - istep
                prevj = j - jstep
                if previ >= 0 and prevj >= 0:
                    cost = (
                        D[previ, prevj]
                        + pwD[i - 1, j - 1] * directional_weights[directionsidx]
                    )
                    if cost < mincost:
                        mincost = cost
                        minidx = directionsidx
                        bestiprev = previ
                        bestjprev = prevj

            D[i, j] = (
                D[bestiprev, bestjprev]
                + pwD[i - 1, j - 1] * directional_weights[minidx]
            )
            B[i - 1, j - 1] = minidx

    # return (dtwd[1:, 1:])
    n = N - 1
    m = M - 1
    step = [m, n]
    path = [step]
    # initialize boolean variables for stopping decoding
    crit = True
    while crit:
        if n == 0 and m == 0:
            crit = False
        else:
            backtracking_pointer = B[m, n]
            bt_vector = directions[backtracking_pointer]
            m -= bt_vector[0]
            n -= bt_vector[1]
            step = [m, n]
        # append next step to the path
        path.append(step)

    output_path = np.array(path, dtype=np.int32)[::-1]
    output_D = D[1:, 1:]
    return output_D, output_path[1:, :]


def dtw_backtracking(dtwd):
    """
    Decode path from the accumulated dtw cost matrix.

    Parameters
    ----------
    dtwd : np.ndarray
        Accumulated cost matrix (computed with
        `dtw_dmatrix_from_pairwise_dmatrix`)

    Returns
    -------
    path : np.ndarray
       A 2D array of size (n_steps, 2), where i-th row has elements
       (i_m, i_n) where i_m represents the index in the input array
       and i_n represents the corresponding index in the reference array.
    """

    N = dtwd.shape[0]
    M = dtwd.shape[1]

    n = N - 1
    m = M - 1

    step = [n, m]

    path = [step]

    # Initialize step choices
    choices = np.zeros((3, 2), dtype=int)
    # Initialize a vector for candidate distances
    dtwd_candidates = np.zeros(3, dtype=float)
    # initialize boolean variables for stopping decoding
    crit = True

    while crit:
        if n == 0:
            # next point in the path
            m = m - 1

        elif m == 0:
            # next point in the path
            n = n - 1

        else:
            # step sizes
            choices[0, 0] = n - 1
            choices[0, 1] = m - 1
            choices[1, 0] = n - 1
            choices[1, 1] = m
            choices[2, 0] = n
            choices[2, 1] = m - 1

            # accumulated distance from the previous step
            # to the next
            dtwd_candidates[0] = dtwd[n - 1, m - 1]
            dtwd_candidates[1] = dtwd[n - 1, m]
            dtwd_candidates[2] = dtwd[n, m - 1]

            # select the best candidate
            p_l_i = np.argmin(dtwd_candidates)

            # update next indices
            n = choices[p_l_i, 0]
            m = choices[p_l_i, 1]

        step = [n, m]
        # append next step to the path
        path.append(step)

        if n == 0 and m == 0:
            crit = False

    return np.array(path[::-1], dtype=int)


def dtw_dmatrix_from_pairwise_dmatrix(D):
    """
    compute dynamic time warping cost matrix
    from a pairwise distance matrix

    Parameters
    ----------
    D : double array
        Pairwise distance matrix (computed e.g., with `cdist`).

    Returns
    -------
    dtwd : np.ndarray
        Accumulated cost matrix
    """
    # Initialize arrays and helper variables
    M = D.shape[0]
    N = D.shape[1]
    # the dtwd distance matrix is initialized with INFINITY
    dtwd = np.ones((M + 1, N + 1), dtype=float) * np.inf

    # Compute the distance iteratively
    dtwd[0, 0] = 0
    for i in range(1, M + 1):
        for j in range(1, N + 1):
            c = D[i - 1, j - 1]
            insertion = dtwd[i - 1, j]
            match = dtwd[i - 1, j - 1]
            deletion = dtwd[i, j - 1]
            dtwd[i, j] = c + min((insertion, deletion, match))

    return dtwd[1:, 1:]


def cdist_dtw_single_loop(arr1, arr2, metric):
    """

    compute  a pairwise distance matrix
    and its dynamic time warping cost matrix

    Parameters
    ----------

    arr1: numpy nd array or list

    arr2: numpy nd array or list

    metric> callable
        a metric function

    Returns
    -------
    dtwd : np.ndarray
        Accumulated cost matrix
    """
    # Initialize arrays and helper variables
    M = len(arr1)  # arr1.shape[0]
    N = len(arr2)  # arr2.shape[0]

    # pdist_array = np.ones((M,N))*np.inf
    # the dtwd distance matrix is initialized with INFINITY
    dtwd = np.ones((M + 1, N + 1), dtype=float) * np.inf

    # Compute the distance iteratively
    dtwd[0, 0] = 0
    for i in range(1, M + 1):
        for j in range(1, N + 1):
            # pdist_array[i-1, j-1] = metric(arr1[i-1], arr2[j-1])
            # c = pdist_array[i - 1, j - 1]
            c = metric(arr1[i - 1], arr2[j - 1])
            insertion = dtwd[i - 1, j]
            deletion = dtwd[i, j - 1]
            match = dtwd[i - 1, j - 1]
            dtwd[i, j] = c + min((insertion, deletion, match))

    return dtwd[1:, 1:]  # pdist_array


# FDTW fw + bw


@jit(nopython=True)
def flexdtw_forward_and_backward(
    pwD,
    directional_weights=np.array([1, 1, 1]),
    directions=np.array([[1, 0], [1, 1], [0, 1]]),
    buffer=1,
):
    """
    compute felxDTW cost matrix,
    backtrace matrix,
    and starting point matrix
    from a pairwise distance matrix

    Parameters
    ----------
    pwD : double array
        Pairwise distance matrix (computed e.g., with `cdist`).
    directional_weights : double array
        weights for each direction
    directions : double array
        directions.
    buffer : int
        buffer for candidate end points

    Returns
    -------
    D : np.ndarray
        Accumulated cost matrix
    B : np.ndarray
        backtrace matrix
    S : np.ndarray
        starting point matrix
    """
    # Initialize arrays and helper variables
    M = pwD.shape[0]
    N = pwD.shape[1]
    D = np.zeros((M, N))  # * np.inf
    B = np.zeros((M, N), dtype=np.int8)  # * -1
    S = np.zeros((M, N), dtype=np.int32)  # * -1
    # initialize matrices
    M_idx = np.arange(M)
    N_idx = np.arange(N)
    D[0, :] = pwD[0, :]
    D[:, 0] = pwD[:, 0]
    S[:, 0] = M_idx
    S[0, :] = -N_idx
    if buffer > N - 2 or buffer > M - 2:
        buffer = min((N - 2, M - 2))
        # raise ValueError("buffer size needs to be smaller than matrix dimensions")

    # Compute the distance iteratively
    for i in range(1, M):
        for j in range(1, N):
            mincost = np.inf
            minidx = -1
            bestiprev = -1
            bestjprev = -1
            for directionsidx, direction in enumerate(directions):
                (istep, jstep) = direction
                previ = i - istep
                prevj = j - jstep
                if previ >= 0 and prevj >= 0:
                    cost = (
                        D[previ, prevj] + pwD[i, j] * directional_weights[directionsidx]
                    )
                    if S[previ, previ] >= 0:
                        dist = i + (j - S[previ, prevj])
                    else:
                        dist = i + (j + S[previ, prevj])
                    cost_per_mb = cost / dist

                    if cost_per_mb < mincost:
                        mincost = cost_per_mb
                        minidx = directionsidx
                        bestiprev = previ
                        bestjprev = prevj

            D[i, j] = D[bestiprev, bestjprev] + pwD[i, j] * directional_weights[minidx]
            B[i, j] = minidx
            S[i, j] = S[bestiprev, bestjprev]

    # get end point
    endpoint_candidates_m = np.column_stack(
        (
            np.full(N - buffer - 1, M - 1, dtype=np.int32),
            np.arange(buffer, N - 1, dtype=np.int32),
        )
    )  # bottom row
    endpoint_candidates_n = np.column_stack(
        (
            np.arange(buffer, M - 1, dtype=np.int32),
            np.full(M - buffer - 1, N - 1, dtype=np.int32),
        )
    )  # right column

    ep_c = np.concatenate(
        (
            endpoint_candidates_m,
            np.array([[M - 1, N - 1]], dtype=np.int32),
            endpoint_candidates_n,
        )
    )
    # endpoints_values = D[ep_c[:,0],ep_c[:,1]] / (np.sum(ep_c, axis = 1) - np.abs(S[ep_c[:,0],ep_c[:,1]]))
    endpoints_values = np.zeros(ep_c.shape[0])
    for idx, ep_c_cand in enumerate(ep_c):
        ep_c1 = ep_c_cand[0]
        ep_c2 = ep_c_cand[1]
        endpoints_values[idx] = D[ep_c1, ep_c2] / (
            ep_c1 + ep_c2 - np.abs(S[ep_c1, ep_c2])
        )

    minimal_ep = np.argmin(endpoints_values)
    m = ep_c[minimal_ep, 0]
    n = ep_c[minimal_ep, 1]
    step = (m, n)
    path = []
    path.append(step)
    # loop over backtracking matrix
    crit = True
    while crit:
        if n == 0 or m == 0:
            crit = False
        else:
            backtracking_pointer = B[m, n]
            bt_vector = directions[backtracking_pointer, :]
            m -= bt_vector[0]
            n -= bt_vector[1]
            step = (m, n)
        # append next step to the path
        path.append(step)
    output_path = np.array(path, dtype=np.int32)[::-1]
    return output_path[1:, :], D, B, S
    # return 1,2,3,4


def flexdtw_dmatrix_from_pairwise_dmatrix(pwD, directional_weights=np.array([1, 1, 1])):
    """
    compute felxDTW cost matrix,
    backtrace matrix,
    and starting point matrix
    from a pairwise distance matrix

    Parameters
    ----------
    pwD : double array
        Pairwise distance matrix (computed e.g., with `cdist`).

    Returns
    -------
    D : np.ndarray
        Accumulated cost matrix
    B : np.ndarray
        backtrace matrix
    S : np.ndarray
        starting point matrix
    """
    # Initialize arrays and helper variables
    M = pwD.shape[0]
    N = pwD.shape[1]
    dw = directional_weights
    D = np.ones((M, N), dtype=float) * np.inf
    B = np.zeros((M, N), dtype=np.int8)  # * -1
    S = np.zeros((M, N), dtype=np.int32)  # * -1
    # initialize matrices
    M_idx = np.arange(M)
    N_idx = np.arange(N)
    D[:, 0] = pwD[:, 0]
    D[0, :] = pwD[0, :]
    S[:, 0] = M_idx
    S[0, :] = -N_idx
    # Compute the distance iteratively
    for i in range(1, M):
        for j in range(1, N):
            c = pwD[i, j]
            S_local = np.array([S[i - 1, j], S[i - 1, j - 1], S[i, j - 1]])
            D_local = np.array(
                [
                    dw[0] * c + D[i - 1, j],
                    dw[1] * c + D[i - 1, j - 1],
                    dw[2] * c + D[i, j - 1],
                ]
            )
            B_local = D_local / (i + j - np.abs(S_local))
            B_dir = np.argmin(B_local)
            B[i, j] = B_dir
            D[i, j] = D_local[B_dir]
            S[i, j] = S_local[B_dir]
    return D, B, S


def flexdtw_backtracking(D, B, S, buffer=1):
    """
    Decode path from the accumulated dtw cost matrix,
    backtrace matrix, and starting point matrix

    Parameters
    ----------
    D : np.ndarray
        Accumulated cost matrix
    B : np.ndarray
        backtrace matrix
    S : np.ndarray
        starting point matrix

    Returns
    -------
    path : np.ndarray
       A 2D array of size (n_steps, 2), where i-th row has elements
       (i_m, i_n) where i_m represents the index in the input array
       and i_n represents the corresponding index in the reference array.
    """

    N = D.shape[0]
    M = D.shape[1]
    if buffer > N - 2 or buffer > M - 2:
        raise ValueError("buffer size needs to be smaller than matrix dimensions")

    endpoint_candidates_n = np.column_stack(
        (np.arange(buffer, N - 1), np.full(N - buffer - 1, M - 1))
    )  # right column
    endpoint_candidates_m = np.column_stack(
        (np.full(M - buffer - 1, N - 1), np.arange(buffer, M - 1))
    )  # bottom row
    ep_c = np.concatenate(
        (endpoint_candidates_n, np.array([[N - 1, M - 1]]), endpoint_candidates_m)
    )

    endpoints_values = D[ep_c[:, 0], ep_c[:, 1]] / (
        np.sum(ep_c, axis=1) - np.abs(S[ep_c[:, 0], ep_c[:, 1]])
    )
    minimal_ep = np.argmin(endpoints_values)
    n = ep_c[minimal_ep, 0]
    m = ep_c[minimal_ep, 1]
    step = np.array([n, m])
    path = [step]

    # initialize boolean variables for stopping decoding
    crit = True
    backtracking_vectors = np.array([[-1, 0], [-1, -1], [0, -1]])

    while crit:
        if step[0] == 0 or step[1] == 0:
            crit = False

        else:
            backtracking_pointer = B[step[0], step[1]]
            bt_vector = backtracking_vectors[backtracking_pointer, :]
            step = np.copy(step) + bt_vector
            path.append(step)

    return np.array(path[::-1], dtype=int)


if __name__ == "__main__":
    A = np.array([[1, 2, 3, 4, 1, 2, 3, 4, 5, 6]]).T
    B = np.array([[1, 2, 3, 4, 5, 6]]).T
    dtwmatcher = WDTW()
    p = dtwmatcher(A, B)
